import os
from typing import Any

import pandas as pd
import requests
from mindsdb_sql import parse_sql
from mindsdb_sql.parser import ast

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.libs.response import HandlerResponse, HandlerStatusResponse
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities.config import Config


class PirateWeatherAPIBaseTable(APITable):
    allowed_select_keys = {}
    columns = []
    table_name = ""

    def __init__(self, handler: "PirateWeatherAPIHandler"):
        super().__init__(handler)

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Select data from the collected weather data and return it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """
        conditions = extract_comparison_conditions(query.where)

        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 in self.allowed_select_keys:
                if op == "=":
                    params[arg1] = arg2
                else:
                    raise NotImplementedError(f"Operator for argument {arg1} is not supported: {op}")

        if "latitude" not in params or "longitude" not in params:
            raise ValueError("Latitude and longitude are required")

        result = self.handler.call_application_api(method_name=self.table_name, params=params)

        # Reparse the query and run through SELECTQueryExecutor
        query_parser = SELECTQueryParser(query, table=self.table_name, columns=self.get_columns())
        selected_columns, where_conditions, order_by_conditions, result_limit = query_parser.parse_query()

        # Remove request parameters from where conditions
        where_conditions = [c for c in where_conditions if c[1] not in self.allowed_select_keys]

        query_executor = SELECTQueryExecutor(result, selected_columns, where_conditions, order_by_conditions,
                                             result_limit)

        return query_executor.execute_query()

    def get_columns(self) -> list:
        return self.columns


class PiratePirateWeatherAPIHourlyTable(PirateWeatherAPIBaseTable):
    allowed_select_keys = {
        "latitude",
        "longitude",
        "time",
        "units"
    }
    columns = [
        "localtime",
        "icon",
        "summary",
        "precipAccumulation",
        "precipType",
        "temperature",
        "apparentTemperature",
        "dewPoint",
        "pressure",
        "windSpeed",
        "windBearing",
        "cloudCover",
        "latitude",
        "longitude",
        "timezone",
        "offset"
    ]
    table_name = "hourly"


class PiratePirateWeatherAPIDailyTable(PirateWeatherAPIBaseTable):
    allowed_select_keys = {
        "latitude",
        "longitude",
        "time",
        "units"
    }

    columns = [
        "localtime",
        "icon",
        "summary",
        "sunriseTime",
        "sunsetTime",
        "moonPhase",
        "precipAccumulation",
        "precipType",
        "temperatureHigh",
        "temperatureHighTime",
        "temperatureLow",
        "temperatureLowTime",
        "apparentTemperatureHigh",
        "apparentTemperatureHighTime",
        "apparentTemperatureLow",
        "apparentTemperatureLowTime",
        "dewPoint",
        "pressure",
        "windSpeed",
        "windBearing",
        "cloudCover",
        "temperatureMin",
        "temperatureMinTime",
        "temperatureMax",
        "temperatureMaxTime",
        "apparentTemperatureMin",
        "apparentTemperatureMinTime",
        "apparentTemperatureMax",
        "apparentTemperatureMaxTime",
        "latitude",
        "longitude",
        "timezone",
        "offset"
    ]
    table_name = "daily"


class PirateWeatherAPIHandler(APIHandler):
    query_string_template = "https://timemachine.pirateweather.net/forecast/{api_key}/{latitude},{longitude}"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)
        self._tables = {}

        args = kwargs.get("connection_data", {})
        handler_config = Config().get("weather_handler", {})

        connection_args = {}

        for k in ["api_key"]:
            if k in args:
                connection_args[k] = args[k]
            elif f"WEATHER_{k.upper()}" in os.environ:
                connection_args[k] = os.environ[f"WEATHER_{k.upper()}"]
            elif k in handler_config:
                connection_args[k] = handler_config[k]

        self._api_key = connection_args["api_key"]

        # Register tables
        self._register_table("hourly", PiratePirateWeatherAPIHourlyTable(self))
        self._register_table("daily", PiratePirateWeatherAPIDailyTable(self))

    def _register_table(self, table_name: str, table_class: Any):
        self._tables[table_name] = table_class

    def connect(self) -> HandlerStatusResponse:
        return HandlerStatusResponse(success=True)

    def check_connection(self) -> HandlerStatusResponse:
        response = HandlerStatusResponse(False)

        try:
            self.call_application_api(method_name="daily", params=dict(latitude=51.507351,
                                                                       longitude=-0.127758,
                                                                       time="1672578052"))
            response.success = True

        except Exception as e:
            response.error_message = str(e)

        return response

    def native_query(self, query: Any):
        ast = parse_sql(query, dialect="mindsdb")
        table = str(ast.from_table)
        data = self._tables[table].select(ast)
        return HandlerResponse(RESPONSE_TYPE.TABLE, data_frame=data)

    def call_application_api(
            self, method_name: str = None, params: dict = None
    ) -> pd.DataFrame:
        # This will implement api base on the native query
        # By processing native query to convert it to api callable parameters
        if method_name not in ["hourly", "daily"]:
            raise NotImplementedError(f"Method {method_name} is not implemented")

        if "latitude" not in params or "longitude" not in params:
            raise ValueError("Latitude and longitude are required")

        opt_params = {
            "exclude": "currently,minutely,alerts,hourly,daily".replace("," + method_name, ""),
            "units": params.get("units"),
        }

        # Build the query
        query = self.query_string_template.format(
            api_key=self._api_key,
            latitude=params["latitude"],
            longitude=params["longitude"]
        )
        if "time" in params:
            query += f",{params['time']}"
        # Add optional parameters
        query += "?" + "&".join([f"{k}={v}" for k, v in opt_params.items() if v])

        # Call the API
        response = requests.get(query)
        response.raise_for_status()

        # Parse the response
        data = response.json()
        if method_name not in data:
            raise ValueError(f"API response did not contain {method_name} data. Check your API key. Got response: {data}")

        # Convert to dataframe
        df = pd.DataFrame(data[method_name]["data"]).assign(
            latitude=params["latitude"],
            longitude=params["longitude"],
            timezone=data["timezone"],
            offset=data["offset"]
        )
        df["localtime"] = pd.to_datetime(df["time"], utc=True, unit="s").dt.tz_convert(data["timezone"])
        df.drop(columns="time", inplace=True)
        return df



from collections import OrderedDict
from typing import Optional
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE



import pandas as pd
import IfxPyDbi as I 
from sqlalchemy_informix.ibmdb import  InformixDialect

logger = log.getLogger(__name__)

class InformixHandler(DatabaseHandler):


    name= 'informix'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        ''' Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        '''
        super().__init__(name)
        
        self.kwargs = kwargs
        self.parser = parse_sql
        self.loging_enabled = connection_data['loging_enabled'] if 'loging_enabled' in connection_data else True
        self.server = connection_data['server']
        self.database = connection_data['database']
        self.user = connection_data['user']
        self.password = connection_data['password']
        self.schemaName = connection_data['schema_name'] 
        self.host = connection_data['host']
        self.port = connection_data['port']
        self.connString = (
    'SERVER={0};'
    'DATABASE={1};'
    'HOST={2};'
    'PORT={3};'
    'UID={4};'
    'PWD={5};').format(self.server, self.database, self.host, self.port, self.user, self.password)

        
        

        self.connection = None
        self.is_connected = False
          
    
    def connect(self):
        ''' Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            Connection Object
        '''
        if self.is_connected is True:
            return self.connection

        try:
            self.connection = I.connect(self.connString,'','')
  
            self.is_connected= True
        except Exception as e:
            logger.error(f'Error while connecting to {self.database}, {e}')


        return self.connection


    def disconnect(self):
        ''' Close any existing connections
        Should switch self.is_connected.
        '''
        if self.is_connected is False:
            return
        try:
            self.connection.close()
            self.is_connected = False
        except Exception as e:
            logger.error(f'Error while disconnecting to {self.database}, {e}')

        return 


    def check_connection(self) -> StatusResponse:
        ''' Check connection to the handler
        Returns:
            HandlerStatusResponse
        '''
        responseCode = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            responseCode.success = True
        except Exception as e:
            logger.error(f'Error connecting to database {self.database}, {e}!')
            responseCode.error_message = str(e)
        finally:
            if responseCode.success is True and need_to_close:
                self.disconnect()
            if responseCode.success is False and self.is_connected is True:
                self.is_connected = False

        return responseCode


    def native_query(self, query: str) -> StatusResponse:
        '''Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        '''
        need_to_close = self.is_connected is False
        conn = self.connect()
        cur = conn.cursor()
        try:
            cur.execute(query)
                   
            if cur._result_set_produced :
                result = cur.fetchall() 
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result,
                        columns=[x[0] for x in cur.description]
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
                if self.loging_enabled:
                    self.connection.commit()
        except Exception as e:
            logger.error(f'Error running query: {query} on {self.database}')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
            if self.loging_enabled:
                self.connection.rollback()

            cur.close()


        if need_to_close is True:
            self.disconnect()

        return response

    
    def query(self, query: ASTNode) -> StatusResponse:
        '''Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        '''
        


        renderer = SqlalchemyRender(InformixDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)


    def get_tables(self) -> StatusResponse:
        ''' Return list of entities
        Return list of entities that will be accesible as tables.
        Returns:
            HandlerResponse: shoud have same columns as information_schema.tables
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-tables-table.html)
                Column 'TABLE_NAME' is mandatory, other is optional.
        '''
        self.connect()


        result=self.connection.tables()
        try:
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        [x['TABLE_NAME'] for x in result if x['TABLE_SCHEM']==self.schemaName],
                        columns=['TABLE_NAME']
                        
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
            
        except Exception as e:
            logger.error(f'Error running while getting table {e} on ')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )



        return response

    
    def get_columns(self, table_name: str) -> StatusResponse:
        ''' Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        '''
        
        self.connect()


        result=self.connection.columns(table_name=table_name)
        try:
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        [result[i]['COLUMN_NAME'] for i in range(len(result)) ],
                        columns=['COLUMN_NAME']
                        
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
            
        except Exception as e:
            logger.error(f'Error running while getting table {e} on ')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )



        return response

        





connection_args = OrderedDict(
    server={
        'type': ARG_TYPE.STR,
        'description': '''
            The server name you want to get connected.
        '''
    },
    database={
        'type': ARG_TYPE.STR,
        'description': '''
            The database name to use when connecting with the DB2 server.
        '''
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the DB2 server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the DB2 server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the DB2 server/database.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect DB2 through TCP/IP'
    },
    schema_name={
        'type': ARG_TYPE.STR,
        'description': 'Specify the schema name for showing tables '
    },
    logging_enabled={
        'type': ARG_TYPE.BOOL,
        'description': '''
        Used for COMMIT and ROLLBACK as this command works only for logging enabled database.
        Note: Its optional.
        Default is TRUE
        '''
    }

)

connection_args_example = OrderedDict(
    server='server',
    database='stores_demo',
    user='informix',
    password='in4mix',
    host='127.0.0.1',
    port='9091',
    schema_name='Love'
    
)



from typing import List

import pandas as pd
from mindsdb_sql.parser import ast

from mindsdb.integrations.utilities.handlers.query_utilities import (
    SELECTQueryExecutor,
    SELECTQueryParser,
)
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.utilities.sql_utils import conditions_to_filter


def val_to_string(d, k):
    if k in d:
        d[k] = str(d[k])


def move_under(d, key_contents_to_move, key_to_move_under=None):
    if key_contents_to_move not in d:
        return
    for k, v in d[key_contents_to_move].items():
        if key_to_move_under:
            d[key_to_move_under][k] = v
        else:
            d[k] = v
    del d[key_contents_to_move]


def select_keys(d, keys):
    new_d = {}
    for key in keys:
        new_d[key] = d.get(key, "")
    return new_d


class CustomAPITable(APITable):

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.handler.connect()

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        return [item for item in self.columns if item not in ignore]

    def select(self, query: ast.Select) -> pd.DataFrame:
        raise NotImplementedError()

    def parse_select(self, query: ast.Select, table_name: str):
        select_statement_parser = SELECTQueryParser(query, table_name, self.get_columns())
        self.selected_columns, self.where_conditions, self.order_by_conditions, self.result_limit = select_statement_parser.parse_query()

    def get_where_param(self, query: ast.Select, param: str):
        params = conditions_to_filter(query.where)
        if param not in params:
            raise Exception(f"WHERE condition does not have '{param}' selector")
        return params[param]

    def apply_query_params(self, df, query):
        select_statement_parser = SELECTQueryParser(query, self.name, self.get_columns())
        selected_columns, _, order_by_conditions, result_limit = select_statement_parser.parse_query()
        select_statement_executor = SELECTQueryExecutor(df, selected_columns, [], order_by_conditions, result_limit)
        return select_statement_executor.execute_query()


class UserTable(CustomAPITable):
    name: str = "user"
    columns: List[str] = [
        'userUuid',
        'email',
        'firstName',
        'lastName',
        'organizationUuid',
        'organizationName',
        'organizationCreatedAt',
        'isTrackingAnonymized',
        'isMarketingOptedIn',
        'isSetupComplete',
        'role',
        'isActive',
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = select_keys(self.connection.get_user(), self.columns)
        df = pd.DataFrame.from_records([data])
        return self.apply_query_params(df, query)


class UserAbilityTable(CustomAPITable):
    name: str = "user_ability"
    columns: List[str] = [
        'action',
        'subject',
        'conditions'
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = select_keys(self.connection.get_user(), ["abilityRules"])
        for d in data:
            val_to_string(d, "condition")
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class OrgTable(CustomAPITable):
    name: str = "org"
    columns: List[str] = [
        'organizationUuid',
        'defaultProjectUuid'
        'name',
        'chartColors',
        'needsProject',
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = select_keys(self.connection.get_org(), self.columns)
        val_to_string(data, "chartColors")
        df = pd.DataFrame.from_records([data])
        return self.apply_query_params(df, query)


class OrgProjectsTable(CustomAPITable):
    name: str = "org_projects"
    columns: List[str] = [
        'name',
        'projectUuid',
        'type',
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get_projects()
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class OrgMembersTable(CustomAPITable):
    name: str = "org_members"
    columns: List[str] = [
        'userUuid',
        'firstName',
        'lastName',
        'email',
        'organizationUuid',
        'role',
        'isActive',
        'isInviteExpired',
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        data = self.connection.get_org_members()
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class ProjectTable(CustomAPITable):
    name: str = "project_table"
    columns: List[str] = [
        'organizationUuid',
        'projectUuid',
        'name',
        'type',
        'pinnedListUuid',
        'copiedFromProjectUuid',
        'dbtVersion',
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, 'project_uuid')
        data = select_keys(self.connection.get_project(project_uuid), self.columns)
        df = pd.DataFrame.from_records([data])
        return self.apply_query_params(df, query)


class WarehouseConnectionTable(CustomAPITable):
    name: str = "warehouse_connection"
    columns: List[str] = [
        "role",
        "type",
        "account",
        "database",
        "warehouse",
        "schema",
        "threads",
        "clientSessionKeepAlive",
        "queryTag",
        "accessUrl",
        "startOfWeek",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, 'project_uuid')
        data = select_keys(self.connection.get_project(project_uuid).get("warehouseConnection", {}), self.columns)
        df = pd.DataFrame.from_records([data])
        return self.apply_query_params(df, query)


class DBTConnectionTable(CustomAPITable):
    name: str = "dbt_connection"
    columns: List[str] = [
        "type",
        "target",
        "profiles_dir",
        "project_dir",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        data = select_keys(self.connection.get_project(project_uuid).get("dbtConnection", {}), self.columns)
        df = pd.DataFrame.from_records([data])
        return self.apply_query_params(df, query)


class DBTEnvironmentVarsTable(CustomAPITable):
    name: str = "dbt_env_vars"
    columns: List[str] = [
        "value",
        "key",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        data = self.connection.get_project(project_uuid).get("dbtConnection", {}).get("environment", [])
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class ChartsTable(CustomAPITable):
    name: str = "charts"
    columns: List[str] = [
        "name",
        "organizationUuid",
        "uuid",
        "description",
        "projectUuid",
        "spaceUuid",
        "pinnedListUuid",
        "spaceName",
        "dashboardUuid",
        "dashboardName",
        "chartType",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        data = self.connection.get_charts_in_project(project_uuid)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class SpacesTable(CustomAPITable):
    name: str = "spaces"
    columns: List[str] = [
        "name",
        "organizationUuid",
        "uuid",
        "projectUuid",
        "pinnedListUuid",
        "pinnedListOrder",
        "isPrivate",
        "dashboardCount",
        "chartCount",
        "access",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        data = self.connection.get_spaces_in_project(project_uuid)
        for d in data:
            val_to_string(d, "access")
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class AccessTable(CustomAPITable):
    name: str = "access"
    columns: List[str] = [
        "lastName",
        "firstName",
        "email",
        "role",
        "projectUuid",
        "userUuid",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        data = self.connection.get_project_access_list(project_uuid)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class ValidationTable(CustomAPITable):
    name: str = "validation"
    columns: List[str] = [
        "source",
        "spaceUuid",
        "projectUuid",
        "errorType",
        "error",
        "name",
        "createdAt",
        "validationId",
        "chartName",
        "chartViews",
        "lastUpdatedAt",
        "lastUpdatedBy",
        "fieldName",
        "chartType",
        "chartUuid",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        data = self.connection.get_validation_results(project_uuid)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class DashboardsTable(CustomAPITable):
    name: str = "dashboards"
    columns: List[str] = [
        "name",
        "organizationUuid",
        "uuid",
        "description",
        "updatedAt",
        "projectUuid",
        "spaceUuid",
        "views",
        "firstViewedAt",
        "pinnedListUuid",
        "pinnedListOrder",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        space_uuid = self.get_where_param(query, "space_uuid")
        data = []
        for row in self.connection.get_space(project_uuid, space_uuid).get("dashboards", []):
            data.append(select_keys(row, self.columns))
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class QueriesTable(CustomAPITable):
    name: str = "queries"
    columns: List[str] = [
        "name",
        "uuid",
        "description",
        "updatedAt",
        "spaceUuid",
        "pinnedListUuid",
        "pinnedListOrder",
        "firstViewedAt",
        "views",
        "chartType",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        space_uuid = self.get_where_param(query, "space_uuid")
        data = []
        for row in self.connection.get_space(project_uuid, space_uuid).get("queries", []):
            data.append(select_keys(row, self.columns))
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class ChartHistoryTable(CustomAPITable):
    name: str = "chart_history"
    columns: List[str] = [
        "createdAt",
        "chartUuid",
        "versionUuid",
        "createdBy",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        chart_uuid = self.get_where_param(query, "chart_uuid")
        data = []
        for row in self.connection.get_chart_version_history(chart_uuid):
            d = select_keys(row, self.columns)
            val_to_string(d, "createdBy")
            data.append(d)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class ChartConfigTable(CustomAPITable):
    name: str = "chart_config"
    columns: List[str] = [
        "legendPosition",
        "showLegend",
        "groupSortOverrides",
        "groupValueOptionOverrides",
        "groupColorOverrides",
        "groupLabelOverrides",
        "showPercentage",
        "showValue",
        "valueLabel",
        "isDonut",
        "metricId",
        "groupFieldIds"
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        chart_uuid = self.get_where_param(query, "chart_uuid")
        version_uuid = self.get_where_param(query, "version_uuid")
        raw_data = self.connection.get_chart(chart_uuid, version_uuid).get("chart", {}).get("chart_config", {})
        config_data = raw_data.get("config", {})
        val_to_string(config_data, "groupSortOverrides")
        val_to_string(config_data, "groupValueOptionOverrides")
        val_to_string(config_data, "groupColorOverrides")
        val_to_string(config_data, "groupLabelOverrides")
        val_to_string(config_data, "groupFieldIds")
        config_data = select_keys(config_data, self.columns)
        df = pd.DataFrame.from_records([{**config_data, "type": raw_data.get("type", "")}], columns=self.columns)
        return self.apply_query_params(df, query)


class ChartAdditionalMetricsTable(CustomAPITable):
    name: str = "chart_additional_metrics"
    columns: List[str] = [
        "label",
        "type",
        "description",
        "sql",
        "hidden",
        "round",
        "compact",
        "format",
        "table",
        "name",
        "index",
        "filters",
        "baseDimensionName",
        "uuid",
        "percentile",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        chart_uuid = self.get_where_param(query, "chart_uuid")
        version_uuid = self.get_where_param(query, "version_uuid")
        data = self.connection.get_chart(chart_uuid, version_uuid).get("metricQuery", {}).get("additionalMetrics", [])
        for d in data:
            val_to_string(data, "filters")
            d = select_keys(d, self.columns)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class ChartTableCalculationsTable(CustomAPITable):
    name: str = "chart_table_calculations"
    columns: List[str] = [
        "suffix",
        "prefix",
        "compact",
        "currency",
        "separator",
        "round",
        "type",
        "sql",
        "displayName",
        "name",
        "index",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        chart_uuid = self.get_where_param(query, "chart_uuid")
        version_uuid = self.get_where_param(query, "version_uuid")
        data = self.connection.get_chart(chart_uuid, version_uuid).get("metricQuery", {}).get("tableCalculations", [])
        for d in data:
            move_under(d, "format")
            d = select_keys(d, self.columns)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class SchedulerLogsTable(CustomAPITable):
    name: str = "scheduler_logs"
    columns: List[str] = [
        "details",
        "targetType",
        "target",
        "status",
        "createdAt",
        "scheduledTime",
        "jobGroup",
        "jobId",
        "schedulerUuid",
        "task",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        project_uuid = self.get_where_param(query, "project_uuid")
        data = self.connection.get_scheduler_logs(project_uuid).get("logs", [])
        for d in data:
            val_to_string(d, "details")
            d = select_keys(d, self.columns)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class SchedulerTable(CustomAPITable):
    name: str = "scheduler"
    columns: List[str] = [
        "options",
        "dashboardUuid",
        "savedChartUuid",
        "cron",
        "format",
        "createdBy",
        "updatedAt",
        "createdAt",
        "message",
        "name",
        "schedulerUuid",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        scheduler_uuid = self.get_where_param(query, "scheduler_uuid")
        data = select_keys(self.connection.get_scheduler(scheduler_uuid), self.columns)
        val_to_string(data, "options")
        df = pd.DataFrame.from_records([data], columns=self.columns)
        return self.apply_query_params(df, query)


class SchedulerJobsTable(CustomAPITable):
    name: str = "scheduler_jobs"
    columns: List[str] = [
        "id",
        "date",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        scheduler_uuid = self.get_where_param(query, "scheduler_uuid")
        data = self.connection.get_scheduler_jobs(scheduler_uuid)
        for d in data:
            d = select_keys(d, self.columns)
        df = pd.DataFrame.from_records(data, columns=self.columns)
        return self.apply_query_params(df, query)


class SchedulerJobStatus(CustomAPITable):
    name: str = "scheduler_job_status"
    columns: List[str] = [
        "status",
    ]

    def __init__(self, handler: APIHandler):
        super().__init__(handler)
        self.connection = self.handler.connect()

    def select(self, query: ast.Select) -> pd.DataFrame:
        job_id = self.get_where_param(query, "job_id")
        data = self.connection.get_scheduler_jobs(job_id)
        data = select_keys(data, self.columns)
        df = pd.DataFrame.from_records([data], columns=self.columns)
        return self.apply_query_params(df, query)



from http import HTTPStatus

from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.storage import db


def create_agent(project_name, name, agent):
    if name is None:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            'Missing field',
            'Missing "name" field for agent'
        )

    if 'model_name' not in agent:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            'Missing field',
            'Missing "model_name" field for agent'
        )

    model_name = agent['model_name']
    params = agent.get('params', {})
    skills = agent.get('skills', [])

    session = SessionController()

    try:
        existing_agent = session.agents_controller.get_agent(name, project_name=project_name)
    except ValueError:
        # Project must exist.
        return http_error(
            HTTPStatus.NOT_FOUND,
            'Project not found',
            f'Project with name {project_name} does not exist'
        )
    if existing_agent is not None:
        return http_error(
            HTTPStatus.CONFLICT,
            'Agent already exists',
            f'Agent with name {name} already exists. Please choose a different one.'
        )

    try:
        created_agent = session.agents_controller.add_agent(
            name,
            project_name,
            model_name,
            skills,
            params=params
        )
        return created_agent.as_dict(), HTTPStatus.CREATED
    except ValueError as e:
        # Model or skill doesn't exist.
        return http_error(
            HTTPStatus.NOT_FOUND,
            'Resource not found',
            str(e)
        )
    except NotImplementedError as e:
        # Free users trying to create agent.
        return http_error(
            HTTPStatus.UNAUTHORIZED,
            'Unavailable to free users',
            str(e)
        )


@ns_conf.route('/<project_name>/agents')
class AgentsResource(Resource):
    @ns_conf.doc('list_agents')
    @api_endpoint_metrics('GET', '/agents')
    def get(self, project_name):
        ''' List all agents '''
        session = SessionController()
        try:
            all_agents = session.agents_controller.get_agents(project_name)
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist')
        return [a.as_dict() for a in all_agents]

    @ns_conf.doc('create_agent')
    @api_endpoint_metrics('POST', '/agents')
    def post(self, project_name):
        '''Create a agent'''

        # Check for required parameters.
        if 'agent' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "agent" parameter in POST body'
            )

        agent = request.json['agent']

        name = agent.get('name')
        return create_agent(project_name, name, agent)


@ns_conf.route('/<project_name>/agents/<agent_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('agent_name', 'Name of the agent')
class AgentResource(Resource):
    @ns_conf.doc('get_agent')
    @api_endpoint_metrics('GET', '/agents/agent')
    def get(self, project_name, agent_name):
        '''Gets a agent by name'''
        session = SessionController()
        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} does not exist'
                )
            return existing_agent.as_dict()
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

    @ns_conf.doc('update_agent')
    @api_endpoint_metrics('PUT', '/agents/agent')
    def put(self, project_name, agent_name):
        '''Updates a agent by name, creating one if it doesn't exist'''

        # Check for required parameters.
        if 'agent' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "agent" parameter in POST body'
            )
        session = SessionController()

        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
        except ValueError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        agent = request.json['agent']
        name = agent.get('name', None)
        model_name = agent.get('model_name', None)
        skills_to_add = agent.get('skills_to_add', [])
        skills_to_remove = agent.get('skills_to_remove', [])
        params = agent.get('params', None)

        # Model needs to exist.
        if model_name is not None:
            session_controller = SessionController()
            model_name_no_version, version = db.Predictor.get_name_and_version(model_name)
            try:
                session_controller.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
            except PredictorRecordNotFound:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Model not found',
                    f'Model with name {model_name} not found'
                )

        # Agent must not exist with new name.
        if name is not None and name != agent_name:
            agent_with_new_name = session.agents_controller.get_agent(name, project_name=project_name)
            if agent_with_new_name is not None:
                return http_error(
                    HTTPStatus.CONFLICT,
                    'Agent already exists',
                    f'Agent with name {name} already exists. Please choose a different one.'
                )

        if existing_agent is None:
            # Create
            return create_agent(project_name, name, agent)

        # Update
        try:
            updated_agent = session.agents_controller.update_agent(
                agent_name,
                project_name=project_name,
                name=name,
                model_name=model_name,
                skills_to_add=skills_to_add,
                skills_to_remove=skills_to_remove,
                params=params
            )
            return updated_agent.as_dict()
        except ValueError as e:
            # Model or skill doesn't exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Resource not found',
                str(e)
            )

    @ns_conf.doc('delete_agent')
    @api_endpoint_metrics('DELETE', '/agents/agent')
    def delete(self, project_name, agent_name):
        '''Deletes a agent by name'''
        session = SessionController()
        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} does not exist'
                )
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        session.agents_controller.delete_agent(agent_name, project_name=project_name)
        return '', HTTPStatus.NO_CONTENT


@ns_conf.route('/<project_name>/agents/<agent_name>/completions')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('agent_name', 'Name of the agent')
class AgentCompletions(Resource):
    @ns_conf.doc('agent_completions')
    @api_endpoint_metrics('POST', '/agents/agent/completions')
    def post(self, project_name, agent_name):
        '''Queries an agent given a list of messages'''
        # Check for required parameters.
        if 'messages' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "messages" parameter in POST body'
            )
        session = SessionController()
        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} does not exist'
                )
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        # Model needs to exist.
        model_name_no_version, version = db.Predictor.get_name_and_version(existing_agent.model_name)
        try:
            agent_model = session.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
            agent_model_record = db.Predictor.query.get(agent_model['id'])
        except PredictorRecordNotFound:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Model not found',
                f'Model with name {existing_agent.model_name} not found'
            )

        completion = session.agents_controller.get_completion(
            existing_agent,
            request.json['messages'],
            project_name=project_name,
            # Don't need to include backoffice_db related tools into this endpoint.
            # Underlying handler (e.g. Langchain) will handle default tools like mdb_read, mdb_write, etc.
            tools=[]
        )
        output_col = agent_model_record.to_predict[0]
        model_output = completion.iloc[-1][output_col]
        return {
            'message': {
                'content': model_output,
                'role': 'assistant'
            }
        }




import pandas as pd
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.api.executor.datahub.datanodes.datanode import DataNode
from mindsdb.api.executor.datahub.datanodes.integration_datanode import (
    IntegrationDataNode,
)
from mindsdb.api.executor.datahub.datanodes.project_datanode import (
    ProjectDataNode,
)
from mindsdb.api.executor import exceptions as exc
from mindsdb.api.executor.utilities.sql import query_df
from mindsdb.api.executor.utilities.sql import get_query_tables
from mindsdb.interfaces.database.projects import ProjectController

from mindsdb.utilities import log

from .system_tables import (
    SchemataTable, TablesTable, ColumnsTable, EventsTable, RoutinesTable,
    PluginsTable, EnginesTable, KeyColumnUsageTable, StatisticsTable,
    CharacterSetsTable, CollationsTable)
from .mindsdb_tables import (
    ModelsTable, DatabasesTable, MLEnginesTable, HandlersTable, JobsTable,
    ChatbotsTable, KBTable, SkillsTable, AgentsTable, ViewsTable, TriggersTable)


logger = log.getLogger(__name__)


class InformationSchemaDataNode(DataNode):
    type = "INFORMATION_SCHEMA"

    tables_list = [
        SchemataTable, TablesTable, ColumnsTable, EventsTable, RoutinesTable,
        PluginsTable, EnginesTable, KeyColumnUsageTable, StatisticsTable,
        CharacterSetsTable, CollationsTable,
        ModelsTable, DatabasesTable, MLEnginesTable, HandlersTable, JobsTable,
        ChatbotsTable, KBTable, SkillsTable, AgentsTable, ViewsTable, TriggersTable
    ]

    def __init__(self, session):
        self.session = session
        self.integration_controller = session.integration_controller
        self.project_controller = ProjectController()
        self.database_controller = session.database_controller

        self.persis_datanodes = {
            'log': self.database_controller.logs_db_controller
        }

        databases = self.database_controller.get_dict()
        if "files" in databases:
            self.persis_datanodes["files"] = IntegrationDataNode(
                "files",
                ds_type="file",
                integration_controller=self.session.integration_controller,
            )

        self.tables = {t.name: t for t in self.tables_list}

    def __getitem__(self, key):
        return self.get(key)

    def get(self, name):
        name_lower = name.lower()

        if name_lower == "information_schema":
            return self

        if name_lower == 'log':
            return self.database_controller.get_system_db('log')

        if name_lower in self.persis_datanodes:
            return self.persis_datanodes[name_lower]

        existing_databases_meta = (
            self.database_controller.get_dict()
        )  # filter_type='project'
        database_name = None
        for key in existing_databases_meta:
            if key.lower() == name_lower:
                database_name = key
                break

        if database_name is None:
            return None

        database_meta = existing_databases_meta[database_name]
        if database_meta["type"] == "integration":
            integration = self.integration_controller.get(name=database_name)
            return IntegrationDataNode(
                database_name,
                ds_type=integration["engine"],
                integration_controller=self.session.integration_controller,
            )
        if database_meta["type"] == "project":
            project = self.database_controller.get_project(name=database_name)
            return ProjectDataNode(
                project=project,
                integration_controller=self.session.integration_controller,
                information_schema=self,
            )

        integration_names = self.integration_controller.get_all().keys()
        for integration_name in integration_names:
            if integration_name.lower() == name_lower:
                datasource = self.integration_controller.get(name=integration_name)
                return IntegrationDataNode(
                    integration_name,
                    ds_type=datasource["engine"],
                    integration_controller=self.session.integration_controller,
                )

        return None

    def has_table(self, tableName):
        tn = tableName.upper()
        if tn in self.tables:
            return True
        return False

    def get_table_columns(self, tableName):
        tn = tableName.upper()
        if tn in self.tables:
            return self.tables[tn].columns
        raise exc.TableNotExistError(
            f"Table information_schema.{tableName} does not exists"
        )

    def get_integrations_names(self):
        integration_names = self.integration_controller.get_all().keys()
        # remove files from list to prevent doubling in 'select from INFORMATION_SCHEMA.TABLES'
        return [x.lower() for x in integration_names if x not in ("files",)]

    def get_projects_names(self):
        projects = self.database_controller.get_dict(filter_type="project")
        return [x.lower() for x in projects]

    def get_tables(self):
        return {
            name: table
            for name, table in self.tables.items()
            if table.visible
        }

    def query(self, query: ASTNode, session=None):
        query_tables = [x[1] for x in get_query_tables(query)]

        if len(query_tables) != 1:
            raise exc.BadTableError(
                f"Only one table can be used in query to information_schema: {query}"
            )

        table_name = query_tables[0].upper()

        if table_name not in self.tables:
            raise exc.NotSupportedYet("Information schema: Not implemented.")

        tbl = self.tables[table_name]

        if hasattr(tbl, 'get_data'):
            dataframe = tbl.get_data(query=query, inf_schema=self, session=self.session)
        else:
            dataframe = self._get_empty_table(tbl)
        data = query_df(dataframe, query, session=self.session)

        columns_info = [{"name": k, "type": v} for k, v in data.dtypes.items()]

        return data.to_dict(orient="split")['data'], columns_info

    def _get_empty_table(self, table):
        columns = table.columns
        data = []

        df = pd.DataFrame(data, columns=columns)
        return df



import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes import Responder
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class Responce(Responder):
    when = {"saslStart": helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            payload = query["payload"].decode()
            mechanism = query.get("mechanism", "SCRAM-SHA-1")
            if mechanism == "SCRAM-SHA-1":
                session.init_scram("sha1")
            elif mechanism == "SCRAM-SHA-256":
                session.init_scram("sha256")
            responce = session.scram.process_client_first_message(payload)

            res = {
                "conversationId": 1,
                "done": False,
                "payload": responce.encode(),
                "ok": 1,
            }
        except Exception as e:
            logger.warning(str(e))
            res = {"errmsg": str(e), "ok": 0}
        return res


responder = Responce()



from mindsdb.integrations.libs.base import BaseMLEngine
import pandas as pd
from mindsdb.integrations.handlers.anomaly_detection_handler.utils import (
    train_unsupervised,
    train_supervised,
    train_semisupervised,
)
from joblib import dump, load
from pyod.models.ecod import ECOD  # unsupervised default
from pyod.models.xgbod import XGBOD  # semi-supervised default
from catboost import CatBoostClassifier  # supervised default
from pyod.models.lof import LOF
from pyod.models.knn import KNN
from pyod.models.pca import PCA
from xgboost import XGBClassifier
from sklearn.naive_bayes import GaussianNB
from sklearn.preprocessing import StandardScaler
import numpy as np
import os


MODELS = {
    "supervised": {
        "catboost": CatBoostClassifier(logging_level="Silent"),
        "xgb": XGBClassifier(),
        "nb": GaussianNB(),
    },
    "semi-supervised": {
        "xgbod": XGBOD(estimator_list=[ECOD()]),
    },
    "unsupervised": {
        "ecod": ECOD(),
        "knn": KNN(),
        "pca": PCA(),
        "lof": LOF(),
    },
}


def choose_model_type(training_df, model_type=None, target=None, supervised_threshold=3000):
    """Choose the model type based on the presence of labels and size of the dataset"""
    if model_type is None:
        if target is None:
            model_type = "unsupervised"
        else:
            model_type = "supervised" if len(training_df) > supervised_threshold else "semi-supervised"
    assert model_type in [
        "supervised",
        "semi-supervised",
        "unsupervised",
    ], "model type must be 'supervised', 'semi-supervised', or 'unsupervised'"
    return model_type


def choose_model(df, model_name=None, model_type=None, target=None, supervised_threshold=3000):
    """Choose the best model based on the size of the dataset and the model type"""
    training_df = preprocess_data(df)
    model_type = choose_model_type(training_df, model_type, target, supervised_threshold)
    if model_name is not None:
        assert model_name in MODELS[model_type], f"model name must be one of {list(MODELS[model_type].keys())}"
        model = MODELS[model_type][model_name]
    else:
        model = None
    if model_type == "unsupervised":
        return train_unsupervised(training_df, model=model)
    X_train = training_df.drop(target, axis=1)
    y_train = training_df[target].astype(int)

    if model_type == "supervised":
        return train_supervised(X_train, y_train, model=model)
    elif model_type == "semi-supervised":
        return train_semisupervised(X_train, y_train)  # Only one semi-supervised model available


def anomaly_type_to_model_name(anomaly_type):
    """Choose the best model name based on the anomaly type"""
    assert anomaly_type in [
        "local",
        "global",
        "clustered",
        "dependency",
    ], "anomaly type must be 'local', 'global', 'clustered', or 'dependency'"
    anomaly_type_dict = {
        "local": "lof",
        "global": "knn",
        "clustered": "pca",
        "dependency": "knn",
    }
    return anomaly_type_dict[anomaly_type]


def preprocess_data(df):
    """Preprocess the data by one-hot encoding categorical columns and scaling numeric columns"""
    # one-hot encode categorical columns
    categorical_columns = list(df.select_dtypes(include=["object"]).columns.values)
    df[categorical_columns] = df[categorical_columns].astype("category")
    df[categorical_columns] = df[categorical_columns].apply(lambda x: x.cat.codes)
    df = pd.get_dummies(df, columns=categorical_columns)
    numeric_columns = list(df.select_dtypes(include=["number"]).columns.values)

    scaler = StandardScaler()
    df[numeric_columns] = scaler.fit_transform(df[numeric_columns])
    return df


def get_model_names(using_args):
    """Get the model names from the using_args. Model names is a list of model names to train.
    If the model is not an ensemble, it only contains one model"""
    model_names = anomaly_type_to_model_name(using_args["anomaly_type"]) if "anomaly_type" in using_args else None
    model_names = using_args["model_name"] if "model_name" in using_args else model_names
    model_names = using_args["ensemble_models"] if "ensemble_models" in using_args else model_names
    model_names = [model_names] if model_names is None else model_names
    model_names = [model_names] if type(model_names) is str else model_names
    return model_names


class AnomalyDetectionHandler(BaseMLEngine):
    """Integration with the PyOD and CatBoost libraries for
    anomaly detection. Both supervised and unsupervised.
    """

    name = "anomaly_detection"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True  # makes unsupervised learning work

    def create(self, target, df, args=None):
        """Train a model and save it to the model storage"""
        args = {} if args is None else args
        using_args = args["using"]
        model_type = using_args["type"] if "type" in using_args else None

        model_names = get_model_names(using_args)

        model_save_paths = []
        model_targets = []
        model_class_names = []
        base_path = self.model_storage.folder_get("context")
        for model_name in model_names:
            model = choose_model(df, model_name=model_name, model_type=model_type, target=target)
            this_model_target = "outlier" if target is None else target  # output column name for unsupervised learning
            save_path = "model.joblib" if model_name is None else model_name + ".joblib"
            dump(model, os.path.join(base_path, save_path))
            model_save_paths.append(save_path)
            model_targets.append(this_model_target)
            model_class_names.append(model.__class__.__name__)
        model_args = {"model_path": model_save_paths, "target": model_targets, "model_name": model_class_names}
        self.model_storage.json_set("model_args", model_args)
        self.model_storage.folder_sync("context")

    def predict(self, df, args=None):
        """Load a model from the model storage and use it to make predictions"""
        args = {} if args is None else args
        model_args = self.model_storage.json_get("model_args")
        results_list = []
        if "__mindsdb_row_id" in df.columns:
            df = df.drop("__mindsdb_row_id", axis=1)
        if model_args["target"][0] in df.columns:
            df = df.drop(model_args["target"], axis=1)
        base_path = self.model_storage.folder_get("context")
        for model_path in model_args["model_path"]:
            model = load(os.path.join(base_path, model_path))
            predict_df = preprocess_data(df).astype(float)
            results = model.predict(predict_df)
            results_list.append(results)
        final_results = np.array(results_list).mean(axis=0)
        final_results = np.where(final_results > 0.5, 1, 0)
        return pd.DataFrame({model_args["target"][0]: final_results})

    def describe(self, attribute="model"):
        model_args = self.model_storage.json_get("model_args")
        df = pd.DataFrame({"model_name": [], "target": []})
        if attribute == "model":
            for model_name, target in zip(model_args["model_name"], model_args["target"]):
                df2 = pd.DataFrame({"model_name": model_name, "target": target}, index=[0])
                df = pd.concat([df, df2], ignore_index=True)
            return df
        else:
            raise NotImplementedError(f"attribute {attribute} not implemented")



from concurrent.futures import ThreadPoolExecutor
import contextvars


class ContextThreadPoolExecutor(ThreadPoolExecutor):
    '''Handles copying context variables to threads created by ThreadPoolExecutor'''
    def __init__(self, max_workers=None):
        self.context = contextvars.copy_context()
        # ThreadPoolExecutor does not propagate context to threads by default, so we need a custom initializer.
        super().__init__(max_workers=max_workers, initializer=self._set_child_context)

    def _set_child_context(self):
        for var, value in self.context.items():
            var.set(value)



from typing import Optional

import pandas as pd
from pydruid.db import connect

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base import DatabaseHandler
from pydruid.db.sqlalchemy import DruidDialect

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)


class DruidHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Apache Druid statements.
    """

    name = 'druid'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'druid'

        optional_parameters = ['user', 'password']
        for parameter in optional_parameters:
            if parameter not in connection_data:
                connection_data[parameter] = None

        if 'path' not in connection_data:
            connection_data['path'] = '/druid/v2/sql/'

        if 'scheme' not in connection_data:
            connection_data['scheme'] = 'http'

        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = connect(
            host=self.connection_data['host'],
            port=self.connection_data['port'],
            path=self.connection_data['path'],
            scheme=self.connection_data['scheme'],
            user=self.connection_data['user'],
            password=self.connection_data['password']
        )
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Pinot, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        cursor = connection.cursor()

        try:
            cursor.execute(query)
            result = cursor.fetchall()
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result,
                        columns=[x[0] for x in cursor.description]
                    )
                )
            else:
                connection.commit()
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(f'Error running query: {query} on Pinot!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        cursor.close()
        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender(DruidDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = """
            SELECT *
            FROM INFORMATION_SCHEMA.TABLES
        """
        result = self.native_query(query)
        df = result.data_frame

        df = df[['TABLE_NAME', 'TABLE_TYPE']]
        result.data_frame = df.rename(columns={'TABLE_NAME': 'table_name', 'TABLE_TYPE': 'table_type'})

        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"""
            SELECT *
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE "TABLE_SCHEMA" = 'druid' AND "TABLE_NAME" = '{table_name}'
        """
        result = self.native_query(query)
        df = result.data_frame

        df = df[['COLUMN_NAME', 'DATA_TYPE']]
        result.data_frame = df.rename(columns={'COLUMN_NAME': 'column_name', 'DATA_TYPE': 'data_type'})

        return result



import time
from datetime import timedelta
from typing import List

from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore

from mindsdb.integrations.utilities.rag.loaders.vector_store_loader.vector_store_loader import VectorStoreLoader
from mindsdb.integrations.utilities.rag.settings import VectorStoreConfig

# gpt-3.5-turbo
_DEFAULT_TPM_LIMIT = 60000
_DEFAULT_RATE_LIMIT_INTERVAL = timedelta(seconds=10)
_INITIAL_TOKEN_USAGE = 0


class VectorStoreOperator:
    """
    Encapsulates the logic for adding documents to a vector store with rate limiting.
    """

    def __init__(self,
                 vector_store: VectorStore,
                 embeddings_model: Embeddings,
                 documents: List[Document] = None,
                 vector_store_config: VectorStoreConfig = None,
                 token_per_minute_limit: int = _DEFAULT_TPM_LIMIT,
                 rate_limit_interval: timedelta = _DEFAULT_RATE_LIMIT_INTERVAL,

                 ):

        self.documents = documents
        self.embeddings_model = embeddings_model
        self.token_per_minute_limit = token_per_minute_limit
        self.rate_limit_interval = rate_limit_interval
        self.current_token_usage = _INITIAL_TOKEN_USAGE
        self._vector_store = None
        self.vector_store_config = vector_store_config

        self.verify_vector_store(vector_store, documents)

    def verify_vector_store(self, vector_store, documents):
        if documents:
            self._add_documents_to_store(documents, vector_store)
        elif isinstance(vector_store, VectorStore):
            # checking is it instance or subclass instance
            self._vector_store = vector_store
        elif issubclass(vector_store, VectorStore):
            # if it is subclass instance, then create instance of it using vector_store_config
            self._vector_store = load_vector_store(self.embeddings_model, self.vector_store_config)

    @property
    def vector_store(self):
        return self._vector_store

    @staticmethod
    def _calculate_token_usage(document):
        return len(document.page_content)

    def _rate_limit(self):
        if self.current_token_usage >= self.token_per_minute_limit:
            time.sleep(self.rate_limit_interval.total_seconds())
            self.current_token_usage = _INITIAL_TOKEN_USAGE

    def _update_token_usage(self, document: Document):
        self._rate_limit()
        self.current_token_usage += self._calculate_token_usage(document)

    def _add_document(self, document: Document):
        self._update_token_usage(document)
        self.vector_store.add_documents([document])

    def _add_documents_to_store(self, documents: List[Document], vector_store: VectorStore):
        self._init_vector_store(documents, vector_store)
        self.add_documents(documents)

    def _init_vector_store(self, documents: List[Document], vector_store: VectorStore):
        if len(documents) > 0:
            self._vector_store = vector_store.from_documents(
                documents=[documents[0]], embedding=self.embeddings_model
            )

    def add_documents(self, documents: List[Document]):
        for document in documents:
            self._add_document(document)


def load_vector_store(embeddings_model: Embeddings, config: VectorStoreConfig) -> VectorStore:
    """
    Loads the vector store based on the provided config and embeddings model
    :param embeddings_model:
    :param config:
    :return:
    """
    loader = VectorStoreLoader(embeddings_model=embeddings_model, config=config)
    return loader.load()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    aws_access_key_id={
        'type': ARG_TYPE.STR,
        'description': 'The access key for the AWS account.'
    },
    aws_secret_access_key={
        'type': ARG_TYPE.STR,
        'description': 'The secret key for the AWS account.',
        'secret': True
    },
    region_name={
        'type': ARG_TYPE.STR,
        'description': 'The AWS region where the DynamoDB tables are created.'
    }
)

connection_args_example = OrderedDict(
    aws_access_key_id='PCAQ2LJDOSWLNSQKOCPW',
    aws_secret_access_key='U/VjewPlNopsDmmwItl34r2neyC6WhZpUiip57i',
    region_name='us-east-1'
)



from mindsdb.integrations.handlers.dockerhub_handler.dockerhub_tables import (
    DockerHubRepoImagesSummaryTable,
    DockerHubRepoImagesTable,
    DockerHubRepoTagTable,
    DockerHubRepoTagsTable,
    DockerHubOrgSettingsTable
)
from mindsdb.integrations.handlers.dockerhub_handler.dockerhub import DockerHubClient
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)

from mindsdb.utilities import log
from mindsdb_sql import parse_sql

logger = log.getLogger(__name__)


class DockerHubHandler(APIHandler):
    """The DockerHub handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the DockerHub handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.docker_client = DockerHubClient()
        self.is_connected = False

        repo_images_stats_data = DockerHubRepoImagesSummaryTable(self)
        self._register_table("repo_images_summary", repo_images_stats_data)
        
        repo_images_data = DockerHubRepoImagesTable(self)
        self._register_table("repo_images", repo_images_data)

        repo_tag_details_data = DockerHubRepoTagTable(self)
        self._register_table("repo_tag_details", repo_tag_details_data)
        
        repo_tags_data = DockerHubRepoTagsTable(self)
        self._register_table("repo_tags", repo_tags_data)

        org_settings = DockerHubOrgSettingsTable(self)
        self._register_table("org_settings", org_settings)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """
        resp = StatusResponse(False)
        status = self.docker_client.login(self.connection_data.get("username"), self.connection_data.get("password"))
        if status["code"] != 200:
            resp.success = False
            resp.error_message = status["error"]
            return resp
        self.is_connected = True
        return resp

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            status = self.docker_client.login(self.connection_data.get("username"), self.connection_data.get("password"))
            if status["code"] == 200:
                current_user = self.connection_data.get("username")
                logger.info(f"Authenticated as user {current_user}")
                response.success = True
            else:
                response.success = False
                logger.info("Error connecting to dockerhub. " + status["error"])
                response.error_message = status["error"]
        except Exception as e:
            logger.error(f"Error connecting to DockerHub API: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)



from .model_executor import ModelExecutor
from .types import Function, BotException


class BotExecutor:
    def __init__(self, chat_task, chat_memory):
        self.chat_task = chat_task
        self.chat_memory = chat_memory

    def _get_model(self, model_name):
        return ModelExecutor(self.chat_task, model_name)

    def _prepare_available_functions(self):

        # collecting functions
        functions = []

        back_db_name = self.chat_task.bot_params.get('backoffice_db')
        if back_db_name is not None:
            back_db = self.chat_task.session.integration_controller.get_data_handler(back_db_name)
            if hasattr(back_db, 'back_office_config'):
                back_db_config = back_db.back_office_config()

                for name, description in back_db_config.get('tools', {}).items():
                    functions.append(
                        Function(
                            name=name,
                            description=description,
                            callback=getattr(back_db, name)
                        ))
        return functions

    def _prepare_skills(self):
        skills = []
        if self.chat_task.agent_id is None:
            return skills
        # Check available skills and translate into a tool.
        agent = self.chat_task.session.agents_controller.get_agent_by_id(
            self.chat_task.agent_id,
            project_name=self.chat_task.project_name
        )
        if agent is None:
            return skills
        for skill in agent.skills:
            skills.append(skill)
        return skills

    def process(self):
        # restart of the bot clear previous history
        if self.chat_memory.get_mode() is None:
            self.chat_memory.hide_history(left_count=1)
            self.chat_memory.set_mode('main')

        functions = self._prepare_available_functions()
        skills = self._prepare_skills()

        model_executor = self._get_model(self.chat_task.base_model_name)
        model_output = model_executor.call(self.chat_memory.get_history(), functions, skills)
        return model_output


class MultiModeBotExecutor(BotExecutor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._modes = self.chat_task.bot_params['modes']

    def _get_avail_modes_items(self):
        return [
            f'- code: {key}, description: {value["info"]}'
            for key, value in self._modes.items()
        ]

    def _make_select_mode_prompt(self):
        # select mode tool
        task_items = self._get_avail_modes_items()

        tasks = '\n'.join(task_items)

        prompt = f'You are a helpful assistant and you can help with various types of tasks.' \
                 f'\nAvailable types of tasks:' \
                 f'\n{tasks}' \
                 f'\nUser have to choose task and assistant MUST call select_task function after it'

        return prompt

    def enter_bot_mode(self, functions):
        # choose prompt or model depending on mode
        mode_name = self.chat_memory.get_mode()

        allowed_tools = None

        if mode_name is None:
            # mode in not selected, lets to go to select menu
            model_executor = self._get_model(self.chat_task.base_model_name)
            prompt = self._make_select_mode_prompt()

            model_executor.prompt = prompt

        else:
            # mode is selected
            mode = self._modes.get(mode_name)
            if mode is None:
                # wrong mode
                self.chat_memory.set_mode(None)
                raise BotException(f'Error to use mode: {mode_name}')

            if 'model' in mode:
                # this is model
                model_executor = self._get_model(mode['model'])

            elif 'prompt' in mode:
                # it is just a prompt. let's use a bot model and custom prompt
                model_executor = self._get_model(self.chat_task.base_model_name)
                model_executor.prompt = mode['prompt']

            else:
                raise BotException(f'Mode is not supported: {mode}')

            allowed_tools = mode.get('allowed_tools')

        if allowed_tools is not None:
            functions = [
                fnc
                for fnc in functions
                if fnc.name in allowed_tools
            ]

        return model_executor, functions

    def _mode_switching_function(self, switched_to_mode):
        # add mode tool

        def _select_task(mode_name):
            if mode_name == '':
                self.chat_memory.set_mode(None)
                switched_to_mode.append(None)
                return 'success'

            avail_modes = list(self._modes.keys())
            if mode_name not in avail_modes:
                return f'Error: task is not found. Available tasks: {", ".join(avail_modes)}'
            self.chat_memory.set_mode(mode_name)
            switched_to_mode.append(mode_name)
            return 'success'

        return Function(
            name='select_task',
            callback=_select_task,
            description='Have to be used by assistant to select task. Input is task type.'
                        ' If user want to unselect task input should be empty string.'
                        ' Available tasks: ' + '; '.join(self._get_avail_modes_items())
        )

    def process(self):
        # this list should be changed if mode was switched
        switched_to_mode = []

        functions_all = self._prepare_available_functions()
        skills = self._prepare_skills()

        # Modes handling
        functions_all.append(self._mode_switching_function(switched_to_mode))

        model_executor, functions = self.enter_bot_mode(functions_all)

        # workaround: don't show history if mode is not selected, otherwise bot doesn't decide to change mode
        if self.chat_memory.get_mode() is None:
            self.chat_memory.hide_history(left_count=1)

        model_output = model_executor.call(self.chat_memory.get_history(), functions, skills)

        if len(switched_to_mode) > 0:
            # mode changed:
            # - clear previous history
            # - run once again

            # start conversation only from last message
            self.chat_memory.hide_history(left_count=1)

            model_executor, functions = self.enter_bot_mode(functions_all)

            model_output = model_executor.call(self.chat_memory.get_history(), functions, skills)

        return model_output



from functools import reduce
from openbb_core.app.static.app_factory import create_app

from mindsdb.integrations.handlers.openbb_handler.openbb_tables import create_table_class
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class OpenBBHandler(APIHandler):
    """A class for handling connections and interactions with the OpenBB Platform.

    Attributes:
        PAT (str): OpenBB's personal access token. Sign up here: https://my.openbb.co
        is_connected (bool): Whether or not the user is connected to their OpenBB account.

    """

    def __init__(self, name: str = None, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)
        self.PAT = None

        args = kwargs.get("connection_data", {})
        if "PAT" in args:
            self.PAT = args["PAT"]

        self.is_connected = False

        # Initialize OpenBB
        # pylint: disable=import-outside-toplevel
        from openbb.package.__extensions__ import Extensions
        self.obb = create_app(Extensions)

        for cmd in list(self.obb.coverage.command_model.keys()):

            openbb_params = self.obb.coverage.command_model[cmd]["openbb"]["QueryParams"]
            openbb_data = self.obb.coverage.command_model[cmd]["openbb"]["Data"]

            # Creates the default data retrieval function for the given command
            # e.g. obb.equity.price.historical, obb.equity.fa.income
            # Note: Even though openbb_params just contains the standard fields that are
            # common across vendors users are able to select any of the fields from the vendor
            # as well. However, some of them might have no effect on the data if the vendor
            # doesn't support it. Regardless, the endpoint won't crash.
            table_class = create_table_class(
                params_metadata=openbb_params,
                response_metadata=openbb_data,
                obb_function=reduce(getattr, cmd[1:].split('.'), self.obb),
                func_docs=f"https://docs.openbb.co/platform/reference/{cmd[1:].replace('.', '/')}"
            )
            self._register_table(cmd.replace('.', '_')[1:], table_class(self))

            # Creates the data retrieval function for each provider
            # e.g. obb.equity.price.historical_polygon, obb.equity.price.historical_intrinio
            for provider in list(self.obb.coverage.command_model[cmd].keys()):

                # Skip the openbb provider since we already created it and it will look like obb.equity.price.historical
                if provider == "openbb":
                    continue

                provider_extra_params = self.obb.coverage.command_model[cmd][provider]["QueryParams"]
                combined_params = provider_extra_params.copy()  # create a copy to avoid modifying the original
                combined_params["fields"] = {**openbb_params["fields"], **provider_extra_params["fields"]}  # merge the fields

                provider_extra_data = self.obb.coverage.command_model[cmd][provider]["Data"]
                combined_data = provider_extra_data.copy()  # create a copy to avoid modifying the original
                combined_data["fields"] = {**openbb_data["fields"], **provider_extra_data["fields"]}  # merge the fields

                table_class = create_table_class(
                    params_metadata=combined_params,
                    response_metadata=combined_data,
                    obb_function=reduce(getattr, cmd[1:].split('.'), self.obb),
                    func_docs=f"https://docs.openbb.co/platform/reference/{cmd[1:].replace('.', '/')}",
                    provider=provider
                )
                self._register_table(f"{cmd.replace('.', '_')[1:]}_{provider}", table_class(self))

    def connect(self) -> bool:
        """Connects with OpenBB account through personal access token (PAT).

        Returns none.
        """
        self.is_connected = False
        self.obb.account.login(pat=self.PAT)

        # Check if PAT utilized is valid
        # if obb.user.profile.active:
        self.is_connected = True
        return True

    def check_connection(self) -> StatusResponse:
        """Checks connection to OpenBB accounting by checking the validity of the PAT.

        Returns StatusResponse indicating whether or not the handler is connected.
        """

        response = StatusResponse(False)

        try:
            if self.connect():
                response.success = True

        except Exception as e:
            logger.error(f"Error connecting to OpenBB Platform: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response



from pandas import DataFrame, concat
from snowflake import connector
from snowflake.sqlalchemy import snowdialect

from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class SnowflakeHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Snowflake statements.
    """

    name = 'snowflake'

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.connection_data = kwargs.get('connection_data')
        self.renderer = SqlalchemyRender(snowdialect.dialect)

        self.is_connected = False
        self.connection = None

    def connect(self):
        """
        Establishes a connection to a Snowflake account.

        Raises:
            ValueError: If the required connection parameters are not provided.
            snowflake.connector.errors.Error: If an error occurs while connecting to the Snowflake account.

        Returns:
            snowflake.connector.connection.SnowflakeConnection: A connection object to the Snowflake account.
        """

        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters
        if not all(key in self.connection_data for key in ['account', 'user', 'password', 'database']):
            raise ValueError('Required parameters (account, user, password, database) must be provided.')

        config = {
            'account': self.connection_data.get('account'),
            'user': self.connection_data.get('user'),
            'password': self.connection_data.get('password'),
            'database': self.connection_data.get('database')
        }

        # Optional connection parameters
        optional_params = ['schema', 'warehouse', 'role']
        for param in optional_params:
            if param in self.connection_data:
                config[param] = self.connection_data[param]

        try:
            self.connection = connector.connect(**config)
            self.is_connected = True
            return self.connection
        except connector.errors.Error as e:
            logger.error(f'Error connecting to Snowflake, {e}!')
            raise

    def disconnect(self):
        """
        Closes the connection to the Snowflake account if it's currently open.
        """

        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Snowflake account.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """

        response = StatusResponse(False)
        need_to_close = not self.is_connected

        try:
            connection = self.connect()

            # Execute a simple query to test the connection
            with connection.cursor() as cur:
                cur.execute('select 1;')
            response.success = True
        except (connector.errors.Error, ValueError) as e:
            logger.error(f'Error connecting to Snowflake, {e}!')
            response.error_message = str(e)

        if response.success and need_to_close:
            self.disconnect()

        elif not response.success and self.is_connected:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query on the Snowflake account and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor(connector.DictCursor) as cur:
            try:
                cur.execute(query)
                result = cur.fetchall()
                if result:
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
            except Exception as e:
                logger.error(f"Error running query: {query} on {self.connection_data.get('database')}, {e}!")
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_code=0,
                    error_message=str(e)
                )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """

        query_str = self.renderer.get_string(query, with_failback=True)
        logger.debug(f"Executing SQL query: {query_str}")
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables and views in the current schema of the Snowflake account.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """

        query = "SHOW TABLES;"
        result_tables = self.native_query(query)
        if result_tables.resp_type == RESPONSE_TYPE.TABLE:
            result_tables.data_frame = result_tables.data_frame.rename(columns={'name': 'table_name'})[['table_name']]
        elif result_tables.resp_type == RESPONSE_TYPE.OK:
            result_tables.data_frame = DataFrame(columns=['table_name'])

        query = "SHOW VIEWS;"
        result_views = self.native_query(query)
        if result_views.resp_type == RESPONSE_TYPE.TABLE:
            result_views.data_frame = result_views.data_frame.rename(columns={'name': 'table_name'})[['table_name']]
        elif result_views.resp_type == RESPONSE_TYPE.OK:
            result_views.data_frame = DataFrame(columns=['table_name'])

        result = Response(RESPONSE_TYPE.TABLE)
        result.data_frame = concat([result_tables.data_frame, result_views.data_frame], ignore_index=True)
        return result

    def get_columns(self, table_name) -> Response:
        """
        Retrieves column details for a specified table in the Snowflake account.

        Args:
            table_name (str): The name of the table for which to retrieve column information.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        Raises:
            ValueError: If the 'table_name' is not a valid string.
        """

        if not table_name or not isinstance(table_name, str):
            raise ValueError("Invalid table name provided.")

        query = f"""
            SELECT COLUMN_NAME AS FIELD, DATA_TYPE AS TYPE
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE TABLE_NAME = '{table_name}';
        """
        result = self.native_query(query)
        result.data_frame = result.data_frame.rename(columns={'FIELD': 'Field', 'TYPE': 'Type'})

        return result



from enum import Enum
from typing import Any
import pandas as pd

from mindsdb.api.executor.utilities.sql import query_df
from mindsdb_sql.parser import ast
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.planner.utils import query_traversal


class FilterOperator(Enum):
    """
    Enum for filter operators.
    """

    EQUAL = "="
    NOT_EQUAL = "!="
    LESS_THAN = "<"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN = ">"
    GREATER_THAN_OR_EQUAL = ">="
    IN = "IN"
    NOT_IN = "NOT IN"
    BETWEEN = "BETWEEN"
    NOT_BETWEEN = "NOT BETWEEN"
    LIKE = "LIKE"
    NOT_LIKE = "NOT LIKE"
    IS_NULL = "IS NULL"
    IS_NOT_NULL = "IS NOT NULL"
    IS = "IS"
    IS_NOT = "IS NOT"


class FilterCondition:
    """
    Base class for filter conditions.
    """

    def __init__(self, column: str, op: FilterOperator, value: Any):
        self.column = column
        self.op = op
        self.value = value
        self.applied = False

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, FilterCondition):
            return (
                self.column == __value.column
                and self.op == __value.op
                and self.value == __value.value
            )
        else:
            return False

    def __repr__(self) -> str:
        return f"""
            FilterCondition(
                column={self.column},
                op={self.op},
                value={self.value}
            )
        """


class SortColumn:
    def __init__(self, column: str, ascending: bool = True):
        self.column = column
        self.ascending = ascending


def make_sql_session():
    from mindsdb.api.executor.controllers.session_controller import SessionController

    sql_session = SessionController()
    sql_session.database = 'mindsdb'
    return sql_session


def conditions_to_filter(binary_op: ASTNode):
    conditions = extract_comparison_conditions(binary_op)

    filters = {}
    for op, arg1, arg2 in conditions:
        if op != '=':
            raise NotImplementedError
        filters[arg1] = arg2
    return filters


def extract_comparison_conditions(binary_op: ASTNode):
    '''Extracts all simple comparison conditions that must be true from an AST node.
    Does NOT support 'or' conditions.
    '''
    conditions = []

    def _extract_comparison_conditions(node: ASTNode, **kwargs):
        if isinstance(node, ast.BinaryOperation):
            op = node.op.lower()
            if op == 'and':
                # Want to separate individual conditions, not include 'and' as its own condition.
                return
            elif not isinstance(node.args[0], ast.Identifier):
                # Only support [identifier] =/</>/>=/<=/etc [constant] comparisons.
                raise NotImplementedError(f'Not implemented arg1: {node.args[0]}')

            if isinstance(node.args[1], ast.Constant):
                value = node.args[1].value
            elif isinstance(node.args[1], ast.Tuple):
                value = [i.value for i in node.args[1].items]
            else:
                raise NotImplementedError(f'Not implemented arg2: {node.args[1]}')

            conditions.append([op, node.args[0].parts[-1], value])

    query_traversal(binary_op, _extract_comparison_conditions)
    return conditions


def project_dataframe(df, targets, table_columns):
    '''
        case-insensitive projection
        'select A' and 'select a' return different column case but with the same content
    '''

    columns = []
    df_cols_idx = {
        col.lower(): col
        for col in df.columns
    }
    df_col_rename = {}

    for target in targets:
        if isinstance(target, ast.Star):
            for col in table_columns:
                col_df = df_cols_idx.get(col.lower())
                if col_df is not None:
                    df_col_rename[col_df] = col
                columns.append(col)

            break
        elif isinstance(target, ast.Identifier):
            col = target.parts[-1]
            col_df = df_cols_idx.get(col.lower())
            if col_df is not None:
                if (
                    hasattr(target, 'alias')
                    and isinstance(target.alias, ast.Identifier)
                ):
                    df_col_rename[col_df] = target.alias.parts[0]
                else:
                    df_col_rename[col_df] = col
            columns.append(col)
        else:
            raise NotImplementedError

    if len(df) == 0:
        df = pd.DataFrame([], columns=columns)
    else:
        # add absent columns
        for col in set(columns) & set(df.columns) ^ set(columns):
            df[col] = None

        # filter by columns
        df = df[columns]

    # adapt column names to projection
    if len(df_col_rename) > 0:
        df = df.rename(columns=df_col_rename)
    return df


def filter_dataframe(df: pd.DataFrame, conditions: list):

    # convert list of conditions to ast.
    # assumes that list was got from extract_comparison_conditions
    where_query = None
    for op, arg1, arg2 in conditions:

        item = ast.BinaryOperation(op=op, args=[ast.Identifier(arg1), ast.Constant(arg2)])
        if where_query is None:
            where_query = item
        else:
            where_query = ast.BinaryOperation(op='and', args=[where_query, item])

    query = ast.Select(targets=[ast.Star()], from_table=ast.Identifier('df'), where=where_query)

    return query_df(df, query)


def sort_dataframe(df, order_by: list):
    cols = []
    ascending = []
    for order in order_by:
        if not isinstance(order, ast.OrderBy):
            continue

        col = order.field.parts[-1]
        if col not in df.columns:
            continue

        cols.append(col)
        ascending.append(False if order.direction.lower() == 'desc' else True)
    if len(cols) > 0:
        df = df.sort_values(by=cols, ascending=ascending)
    return df



import traceback
import datetime as dt

from mindsdb.integrations.libs.api_handler import APIChatHandler

from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.tasks.task import BaseTask

from mindsdb.utilities import log

from .polling import MessageCountPolling, RealtimePolling
from .memory import DBMemory, HandlerMemory
from .chatbot_executor import MultiModeBotExecutor, BotExecutor

from .types import ChatBotMessage

logger = log.getLogger(__name__)


class ChatBotTask(BaseTask):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bot_id = self.object_id
        self.agent_id = None

        self.session = SessionController()

    def run(self, stop_event):

        # TODO check deleted, raise errors
        # TODO checks on delete predictor / project/ integration

        bot_record = db.ChatBots.query.get(self.bot_id)

        self.base_model_name = bot_record.model_name
        self.agent_id = bot_record.agent_id
        self.project_name = db.Project.query.get(bot_record.project_id).name
        self.project_datanode = self.session.datahub.get(self.project_name)

        database_name = db.Integration.query.get(bot_record.database_id).name

        self.chat_handler = self.session.integration_controller.get_data_handler(database_name)
        if not isinstance(self.chat_handler, APIChatHandler):
            raise Exception(f"Can't use chat database: {database_name}")

        # get chat handler info
        self.bot_params = bot_record.params or {}

        chat_params = self.chat_handler.get_chat_config()
        self.bot_params['bot_username'] = self.chat_handler.get_my_user_name()

        polling = chat_params['polling']['type']
        if polling == 'message_count':
            self.chat_pooling = MessageCountPolling(self, chat_params)
            self.memory = HandlerMemory(self, chat_params)

        elif polling == 'realtime':
            self.chat_pooling = RealtimePolling(self, chat_params)
            self.memory = DBMemory(self, chat_params)
        else:
            raise Exception(f"Not supported polling: {polling}")

        if self.bot_params.get('modes') is None:
            self.bot_executor_cls = BotExecutor
        else:
            self.bot_executor_cls = MultiModeBotExecutor

        self.chat_pooling.run(stop_event)

    def on_message(self, chat_memory, message: ChatBotMessage):

        try:
            self._on_message(chat_memory, message)
        except (SystemExit, KeyboardInterrupt):
            raise
        except Exception:
            error = traceback.format_exc()
            logger.error(error)
            self.set_error(str(error))

    def _on_message(self, chat_memory, message: ChatBotMessage):
        # add question to history
        # TODO move it to realtime pooling
        chat_memory.add_to_history(message)

        logger.debug(f'>>chatbot {chat_memory.chat_id} in: {message.text}')

        # process
        bot_executor = self.bot_executor_cls(self, chat_memory)
        response_text = bot_executor.process()

        chat_id = chat_memory.chat_id
        bot_username = self.bot_params['bot_username']
        response_message = ChatBotMessage(
            ChatBotMessage.Type.DIRECT,
            response_text,
            # In Slack direct messages are treated as channels themselves.
            user=bot_username,
            destination=chat_id,
            sent_at=dt.datetime.now()
        )

        # send to chat adapter
        self.chat_pooling.send_message(response_message)
        logger.debug(f'>>chatbot {chat_id} out: {response_message.text}')

        # send to history
        chat_memory.add_to_history(response_message)



import ast
import uuid

from langchain_core.embeddings import Embeddings
from langchain_community.vectorstores import Chroma, PGVector
from langchain_core.vectorstores import VectorStore
from pydantic import BaseModel

from mindsdb.integrations.libs.vectordatabase_handler import TableField
from mindsdb.integrations.utilities.rag.settings import VectorStoreType, VectorStoreConfig
from mindsdb.utilities import log

import pandas as pd

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.exc import DisconnectionError

logger = log.getLogger(__name__)


class VectorStoreLoader(BaseModel):
    embeddings_model: Embeddings
    vector_store: VectorStore = None
    config: VectorStoreConfig = None

    class Config:
        arbitrary_types_allowed = True
        extra = "forbid"
        validate_assignment = True

    def load(self) -> VectorStore:
        """
        Loads the vector store based on the provided config and embeddings model
        :return:
        """
        self.vector_store = VectorStoreFactory.create(self.embeddings_model, self.config)
        return self.vector_store


class VectorStoreFactory:
    @staticmethod
    def create(embeddings_model: Embeddings, config: VectorStoreConfig):

        if config.vector_store_type == VectorStoreType.CHROMA:
            return VectorStoreFactory._load_chromadb_store(embeddings_model, config)
        elif config.vector_store_type == VectorStoreType.PGVECTOR:
            return VectorStoreFactory._load_pgvector_store(embeddings_model, config)
        else:
            raise ValueError(f"Invalid vector store type, must be one either {VectorStoreType.__members__.keys()}")

    @staticmethod
    def _load_chromadb_store(embeddings_model: Embeddings, settings) -> Chroma:
        return Chroma(
            persist_directory=settings.persist_directory,
            collection_name=settings.collection_name,
            embedding_function=embeddings_model,
        )

    @staticmethod
    def _load_pgvector_store(embeddings_model: Embeddings, settings) -> PGVector:
        # create an empty store if collection_name does not exist otherwise load the existing collection
        store = PGVector(
            connection_string=settings.connection_string,
            collection_name=settings.collection_name,
            embedding_function=embeddings_model
        )
        return VectorStoreFactory._load_data_into_langchain_pgvector(settings, store)

    @staticmethod
    def _load_data_into_langchain_pgvector(settings, vectorstore: PGVector) -> PGVector:
        """
        Fetches data from the existing pgvector table and loads it into the langchain pgvector vector store
        :param settings:
        :param vectorstore:
        :return:
        """
        df = VectorStoreFactory._fetch_data_from_db(settings)

        df[TableField.EMBEDDINGS] = df[TableField.EMBEDDINGS].apply(ast.literal_eval)
        df[TableField.METADATA] = df[TableField.METADATA].apply(ast.literal_eval)

        metadata = df[TableField.METADATA].tolist()
        embeddings = df[TableField.EMBEDDINGS].tolist()
        texts = df[TableField.CONTENT].tolist()
        ids = [str(uuid.uuid1()) for _ in range(len(df))] \
            if TableField.ID not in df.columns else df[TableField.ID].tolist()

        vectorstore.add_embeddings(
            texts=texts,
            embeddings=embeddings,
            metadatas=metadata,
            ids=ids
        )
        return vectorstore

    @staticmethod
    def _fetch_data_from_db(settings: VectorStoreConfig) -> pd.DataFrame:
        """
        Fetches data from the database using the provided connection_string in the settings
        :param settings:
        :return:
        """
        try:
            engine = create_engine(settings.connection_string)
            db = scoped_session(sessionmaker(bind=engine))

            df = pd.read_sql(f"SELECT * FROM {settings.collection_name}", engine)

            return df
        except DisconnectionError as e:
            logger.error("Unable to connect to the database. Please check your connection string and try again.")
            raise e
        except Exception as e:
            logger.error(f"An error occurred while fetching data from the database: {e}")
            raise e
        finally:
            db.close()



from mindsdb.utilities.ml_task_queue.utils import wait_redis_ping


class BaseRedisQueue:
    def wait_redis_ping(self, timeout: int = 30) -> None:
        """ wait when redis.ping return True

            Args:
                timeout (int): seconds to wait for success ping

            Raises:
                RedisConnectionError: if `ping` did not return `True` within `timeout` seconds
        """
        return wait_redis_ping(self.db, timeout)



from typing import List
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb_sql.parser import ast
import pandas as pd
from mindsdb_sql.parser.ast.select.constant import Constant
import json


class Articles(APITable):
    name: str = "articles"

    def __init__(self, handler: APIHandler):
        super().__init__(handler)

    def select(self, query: ast.Select) -> pd.DataFrame:
        """triggered at the SELECT query

        Args:
            query (ast.Select): user's entered query

        Returns:
            pd.DataFrame: the queried information
        """
        _id = None
        selected_columns = []

        # Get id from where clause, if available
        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise ValueError("Unsupported condition in WHERE clause")

        # Get selected columns from query
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        # Initialize the result DataFrame
        result_df = None

        if _id is not None:
            # Fetch data using the provided endpoint for the specific id
            df = self.handler.call_intercom_api(endpoint=f'/articles/{_id}')

            if len(df) > 0:
                result_df = df[selected_columns]
        else:
            # Fetch data without specifying an id
            page_size = 100  # The page size you want to use for API requests
            limit = query.limit.value if query.limit else None
            result_df = pd.DataFrame(columns=selected_columns)

            if limit:
                # Calculate the number of pages required
                page_count = (limit + page_size - 1) // page_size
            else:
                page_count = 1

            for page in range(1, page_count + 1):
                if limit == 0:
                    break
                if limit:
                    # Calculate the page size for this request
                    current_page_size = min(page_size, limit)
                else:
                    current_page_size = page_size

                df = pd.DataFrame(self.handler.call_intercom_api(endpoint='/articles', params={'page': page, 'per_page': current_page_size})['data'][0])
                if len(df) == 0:
                    break
                result_df = pd.concat([result_df, df[selected_columns]], ignore_index=True)
                if limit:
                    limit -= current_page_size
        return result_df

    def insert(self, query: ast.Insert) -> None:
        """insert

        Args:
            query (ast.Insert): user's entered query

        Returns:
            None
        """
        data = {}
        for column, value in zip(query.columns, query.values[0]):
            if isinstance(value, Constant):
                data[column.name] = value.value
            else:
                data[column.name] = value
        self.handler.call_intercom_api(endpoint='/articles', method='POST', data=json.dumps(data))

    def update(self, query: ast.Update) -> None:
        """update

        Args:
            query (ast.Update): user's entered query

        Returns:
            None
        """
        conditions = extract_comparison_conditions(query.where)
        # Get page id from query
        _id = None
        for op, arg1, arg2 in conditions:
            if arg1 == 'id' and op == '=':
                _id = arg2
            else:
                raise NotImplementedError

        data = {}
        for key, value in query.update_columns.items():
            if isinstance(value, Constant):
                data[key] = value.value
            else:
                data[key] = value
        self.handler.call_intercom_api(endpoint=f'/articles/{_id}', method='PUT', data=json.dumps(data))

    def get_columns(self, ignore: List[str] = []) -> List[str]:
        """columns

        Args:
            ignore (List[str], optional): exclusion items. Defaults to [].

        Returns:
            List[str]: available columns with `ignore` items removed from the list.
        """
        return [
            "type",
            "id",
            "workspace_id",
            "title",
            "description",
            "body",
            "author_id",
            "state",
            "created_at",
            "updated_at",
            "url",
            "parent_id",
            "parent_ids",
            "parent_type",
            "statistics"
        ]



from typing import Optional

from pyignite import Client
import pandas as pd

from mindsdb_sql import parse_sql
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


logger = log.getLogger(__name__)

class IgniteHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Apache Ignite statements.
    """

    name = 'ignite'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'ignite'

        optional_parameters = ['username', 'password', 'schema']
        for parameter in optional_parameters:
            if parameter not in connection_data:
                connection_data[parameter] = None

        self.connection_data = connection_data
        self.kwargs = kwargs

        self.client = None
        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.client = Client(
            username=self.connection_data['username'],
            password=self.connection_data['password']
        )

        try:
            port = int(self.connection_data['port'])
        except ValueError:
            raise ValueError("Invalid port number")

        nodes = [(self.connection_data['host'], port)]
        self.connection = self.client.connect(nodes)
        self.is_connected = True

        return self.client, self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.client.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Apache Ignite!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        client, connection = self.connect()

        try:
            with connection:
                with client.sql(query, include_field_names=True, schema=self.connection_data['schema']) as cursor:
                    result = list(cursor)
                    if result and result[0][0] != 'UPDATED':
                        response = Response(
                            RESPONSE_TYPE.TABLE,
                            data_frame=pd.DataFrame(
                                result[1:],
                                columns=result[0]
                            )
                        )
                    else:
                        response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(f'Error running query: {query} on Apache Ignite!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        cursor.close()
        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        if isinstance(query, ASTNode):
            query_str = query.to_string()
        else:
            query_str = str(query)

        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = """
            SELECT TABLE_NAME FROM SYS.TABLES
        """
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"""
            SELECT COLUMN_NAME, TYPE FROM SYS.TABLE_COLUMNS WHERE TABLE_NAME = '{table_name.upper()}'
        """
        result = self.native_query(query)
        df = result.data_frame
        df['TYPE'] = df.apply(lambda row: row['TYPE'].split('.')[-1], axis=1)
        df = df.iloc[2:]
        result.data_frame = df.rename(columns={'COLUMN_NAME': 'column_name', 'TYPE': 'data_type'})
        return result



import copy
import pandas as pd

from mindsdb.api.executor.exceptions import WrongArgumentError


class Column:
    def __init__(self, name=None, alias=None,
                 table_name=None, table_alias=None,
                 type=None, database=None, flags=None,
                 charset=None):
        if alias is None:
            alias = name
        if table_alias is None:
            table_alias = table_name
        self.name = name
        self.alias = alias
        self.table_name = table_name
        self.table_alias = table_alias
        self.type = type
        self.database = database
        self.flags = flags
        self.charset = charset

    def get_hash_name(self, prefix):
        table_name = self.table_name if self.table_alias is None else self.table_alias
        name = self.name if self.alias is None else self.alias

        name = f'{prefix}_{table_name}_{name}'
        return name

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__})'


class ResultSet:
    def __init__(self, length=0):
        self._columns = []
        # records is list of lists with the same length as columns
        self._records = []
        for i in range(length):
            self._records.append([])

        self.is_prediction = False

    def __repr__(self):
        col_names = ', '.join([col.name for col in self._columns])
        data = '\n'.join([str(rec) for rec in self._records[:20]])

        if len(self._records) > 20:
            data += '\n...'

        return f'{self.__class__.__name__}({self.length()} rows, cols: {col_names})\n {data}'

    def __len__(self) -> int:
        return len(self._records)

    # --- converters ---

    def from_df(self, df, database, table_name, table_alias=None):

        resp_dict = df.to_dict(orient='split')

        self._records = resp_dict['data']

        for col in resp_dict['columns']:
            self._columns.append(Column(
                name=col,
                table_name=table_name,
                table_alias=table_alias,
                database=database,
                type=df.dtypes[col]
            ))
        return self

    def from_df_cols(self, df, col_names, strict=True):
        # find column by alias
        alias_idx = {}
        for col in col_names.values():
            if col.alias is not None:
                alias_idx[col.alias] = col

        resp_dict = df.to_dict(orient='split')

        self._records = resp_dict['data']

        for col in resp_dict['columns']:
            if col in col_names or strict:
                column = col_names[col]
            elif col in alias_idx:
                column = alias_idx[col]
            else:
                column = Column(col)
            self._columns.append(column)
        return self

    def to_df(self):
        columns = self.get_column_names()
        return pd.DataFrame(self._records, columns=columns)

    def to_df_cols(self, prefix=''):
        # returns dataframe and dict of columns
        #   can be restored to ResultSet by from_df_cols method

        columns = []
        col_names = {}
        for col in self._columns:
            name = col.get_hash_name(prefix)
            columns.append(name)
            col_names[name] = col

        return pd.DataFrame(self._records, columns=columns), col_names

    # --- tables ---

    def get_tables(self):
        tables_idx = []
        tables = []
        cols = ['database', 'table_name', 'table_alias']
        for col in self._columns:
            table = (col.database, col.table_name, col.table_alias)
            if table not in tables_idx:
                tables_idx.append(table)
                tables.append(dict(zip(cols, table)))
        return tables

    # --- columns ---

    def _locate_column(self, col):
        col_idx = None
        for i, col0 in enumerate(self._columns):
            if col0 is col:
                col_idx = i
                break
        if col_idx is None:
            raise WrongArgumentError(f'Column is not found: {col}')
        return col_idx

    def add_column(self, col, values=None):
        self._columns.append(col)

        if values is None:
            values = []
        # update records
        if len(self._records) > 0:
            for rec in self._records:
                if len(values) > 0:
                    value = values.pop(0)
                else:
                    value = None
                rec.append(value)

    def del_column(self, col):
        idx = self._locate_column(col)
        self._columns.pop(idx)
        for row in self._records:
            row.pop(idx)

    @property
    def columns(self):
        return self._columns

    def get_column_names(self):
        columns = [
            col.name if col.alias is None else col.alias
            for col in self._columns
        ]
        return columns

    def find_columns(self, alias=None, table_alias=None):
        col_list = []
        for col in self.columns:
            if alias is not None and col.alias.lower() != alias.lower():
                continue
            if table_alias is not None and col.table_alias.lower() != table_alias.lower():
                continue
            col_list.append(col)

        return col_list

    def copy_column_to(self, col, result_set2):
        # copy with values
        idx = self._locate_column(col)

        values = [row[idx] for row in self._records]

        col2 = copy.deepcopy(col)

        result_set2.add_column(col2, values)
        return col2

    # --- records ---

    def add_records(self, data):
        names = self.get_column_names()
        for rec in data:
            # if len(rec) != len(self._columns):
            #     raise ErSqlWrongArguments(f'Record length mismatch columns length: {len(rec)} != {len(self._columns)}')

            record = [
                rec[name]
                for name in names
            ]
            self._records.append(record)

    def get_records_raw(self):
        return self._records

    def add_record_raw(self, rec):
        if len(rec) != len(self._columns):
            raise WrongArgumentError(f'Record length mismatch columns length: {len(rec)} != {len(self.columns)}')
        self._records.append(rec)

    @property
    def records(self):
        return self.get_records()

    def get_records(self):
        # get records as dicts.
        # !!! Attention: !!!
        # if resultSet contents duplicate column name: only one of them will be in output
        names = self.get_column_names()
        records = []
        for row in self._records:
            records.append(dict(zip(names, row)))
        return records

    # def clear_records(self):
    #     self._records = []

    def length(self):
        return len(self._records)



import pathlib
from typing import Iterator, List

from langchain_community.document_loaders.base import BaseLoader
from langchain_community.document_loaders.csv_loader import CSVLoader
from langchain_community.document_loaders import PyMuPDFLoader
from langchain_community.document_loaders import TextLoader
from langchain_community.document_loaders import UnstructuredHTMLLoader
from langchain_community.document_loaders import UnstructuredMarkdownLoader
from langchain_core.documents.base import Document


class FileLoader(BaseLoader):
    '''Loads files of various types into vector database document representation'''
    def __init__(self, path: str):
        self.path = path
        super().__init__()

    def _get_loader_from_extension(self, extension: str, path: str) -> BaseLoader:
        if extension == '.pdf':
            return PyMuPDFLoader(path)
        if extension == '.csv':
            return CSVLoader(path)
        if extension == '.html':
            return UnstructuredHTMLLoader(path)
        if extension == '.md':
            return UnstructuredMarkdownLoader(path)
        return TextLoader(path, encoding='utf-8')

    def _lazy_load_documents_from_file(self, path: str) -> Iterator[Document]:
        file_extension = pathlib.Path(path).suffix
        loader = self._get_loader_from_extension(file_extension, path)

        for doc in loader.lazy_load():
            doc.metadata['extension'] = file_extension
            yield doc

    def load(self) -> List[Document]:
        '''Loads a file and converts the contents into a vector database Document representation'''
        return list(self.lazy_load())

    def lazy_load(self) -> Iterator[Document]:
        '''Loads a file and converts the contents into a vector database Document representation'''
        for doc in self._lazy_load_documents_from_file(self.path):
            yield doc



from collections import OrderedDict
from typing import Optional
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


import pandas as pd
from crate import client as db
from crate.client.sqlalchemy.dialect import CrateDialect

logger = log.getLogger(__name__)


class CrateHandler(DatabaseHandler):
    name = "crate"

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)

        self.kwargs = kwargs
        self.parser = parse_sql
        self.dialect = "crate"
        self.user = connection_data["user"]
        self.password = connection_data["password"]
        self.schemaName = connection_data.get("schema_name", "doc")
        self.host = connection_data["host"]
        self.port = connection_data["port"]

        self.connection = None
        self.is_connected = False

    def connect(self):
        """Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            Connection Object
        """
        if self.is_connected is True:
            return self.connection

        url = "http://{0}:{1}@{2}:{3}".format(
            self.user, self.password, self.host, self.port
        )
        try:
            self.connection = db.connect(url)

            self.is_connected = True
        except Exception as e:
            logger.error(f"Error while connecting to CrateDB, {e}")

        return self.connection

    def disconnect(self):
        """Close any existing connections
        Should switch self.is_connected.
        """

        if self.is_connected is False:
            return
        try:
            self.connection.close()
            self.is_connected = False
        except Exception as e:
            logger.error(f"Error while disconnecting to CrateDB, {e}")

        return

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler
        Returns:
            HandlerStatusResponse
        """

        responseCode = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            responseCode.success = True
        except Exception as e:
            logger.error(f"Error connecting to  CrateDB, {e}!")
            responseCode.error_message = str(e)
        finally:
            if responseCode.success is True and need_to_close:
                self.disconnect()
            if responseCode.success is False and self.is_connected is True:
                self.is_connected = False

        return responseCode

    def native_query(self, query: str) -> StatusResponse:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        conn = self.connect()
        cur = conn.cursor()
        try:
            cur.execute(query)
            if cur.rowcount:
                result = cur.fetchall()
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result, columns=[x[0] for x in cur.description]
                    ),
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(f"Error running query: {query} on CrateDB!")
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))
        cur.close()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        renderer = SqlalchemyRender(CrateDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """Return list of entities
        Return list of entities that will be accesible as tables.
        Returns:
            HandlerResponse: shoud have same columns as information_schema.tables
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-tables-table.html)
                Column 'TABLE_NAME' is mandatory, other is optional.
        """

        q = f"SHOW TABLES FROM {self.schemaName};"
        result = self.native_query(q)
        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """

        q = f"SHOW COLUMNS FROM {table_name};"
        result = self.native_query(q)
        return result


connection_args = OrderedDict(
    host={
        "type": ARG_TYPE.STR,
        "description": "The host name or IP address of the CrateDB server/database.",
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the CrateDB server.",
    },
    password={
        "type": ARG_TYPE.STR,
        "description": "The password to authenticate the user with the CrateDB server.",
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "Specify port to connect CrateDB server",
    },
    schemaName={
        "type": ARG_TYPE.STR,
        "description": 'Specify the schema name. Note: It is optional DEFAULT is "doc"',
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1",
    port="4200",
    password="",
    user="crate",
)



# this dict is used to configure the data loader and match one on one arguments passed by the user in args
data_loaders = {
    "DFReader": {
        # TODO: add parameter of DFReader
    },
    "SimpleWebPageReader": {
        "source_url_link": "<url>",
    },
    "GithubRepositoryReader": {
        "owner": "<owner>",
        "repo": "<repository>",
        "github_token": "<github_token>",
        "branch": "<branch>",
    },
    "YoutubeTranscriptReader": {
        "ytlinks": ["<youtube_link>"],
    },
}



from typing import Optional, Dict
import pandas as pd

from mindsdb.integrations.handlers.clipdrop_handler.clipdrop import ClipdropClient

from mindsdb.integrations.libs.base import BaseMLEngine

from mindsdb.utilities import log

from mindsdb.integrations.utilities.handler_utils import get_api_key


logger = log.getLogger(__name__)


class ClipdropHandler(BaseMLEngine):
    name = "clipdrop"

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        args = args['using']

        available_tasks = ["remove_text", "remove_background", "sketch_to_image", "text_to_image", "replace_background", "reimagine"]

        if 'task' not in args:
            raise Exception('task has to be specified. Available tasks are - ' + available_tasks)

        if args['task'] not in available_tasks:
            raise Exception('Unknown task specified. Available tasks are - ' + available_tasks)

        if 'local_directory_path' not in args:
            raise Exception('local_directory_path has to be specified')

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("Clipdrop AI Inference engine requires a USING clause! Refer to its documentation for more details.")
        self.generative = True

        args = args['using']
        args['target'] = target
        self.model_storage.json_set('args', args)

    def _get_clipdrop_client(self, args):
        api_key = get_api_key('clipdrop', args["using"], self.engine_storage, strict=False)

        local_directory_path = args["local_directory_path"]

        return ClipdropClient(api_key=api_key, local_dir=local_directory_path)

    def _process_remove_text(self, df, args):

        def generate_remove_text(conds, client):
            conds = conds.to_dict()
            return client.remove_text(conds.get("image_url"))

        supported_params = set(["image_url"])

        if "image_url" not in df.columns:
            raise Exception("`image_url` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for remove text - {supported_params}")

        client = self._get_clipdrop_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_remove_text, client=client, axis=1)

    def _process_remove_background(self, df, args):

        def generate_remove_background(conds, client):
            conds = conds.to_dict()
            return client.remove_background(conds.get("image_url"))

        supported_params = set(["image_url"])

        if "image_url" not in df.columns:
            raise Exception("`image_url` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for remove background - {supported_params}")

        client = self._get_clipdrop_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_remove_background, client=client, axis=1)

    def _process_sketch_to_image(self, df, args):

        def generate_sketch_to_image(conds, client):
            conds = conds.to_dict()
            return client.sketch_to_image(conds.get("image_url"), conds.get("text"))

        supported_params = set(["image_url", "text"])

        if "image_url" not in df.columns:
            raise Exception("`image_url` column has to be given in the query.")

        if "text" not in df.columns:
            raise Exception("`text` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for remove background - {supported_params}")

        client = self._get_clipdrop_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_sketch_to_image, client=client, axis=1)

    def _process_text_to_image(self, df, args):

        def generate_text_to_image(conds, client):
            conds = conds.to_dict()
            return client.text_to_image(conds.get("text"))

        supported_params = set(["text"])

        if "text" not in df.columns:
            raise Exception("`text` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for remove background - {supported_params}")

        client = self._get_clipdrop_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_text_to_image, client=client, axis=1)

    def _process_replace_background(self, df, args):

        def generate_replace_background(conds, client):
            conds = conds.to_dict()
            return client.replace_background(conds.get("image_url"), conds.get("text"))

        supported_params = set(["image_url", "text"])

        if "image_url" not in df.columns:
            raise Exception("`image_url` column has to be given in the query.")

        if "text" not in df.columns:
            raise Exception("`text` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for replace background - {supported_params}")

        client = self._get_clipdrop_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_replace_background, client=client, axis=1)

    def _process_reimagine(self, df, args):

        def generate_reimagine(conds, client):
            conds = conds.to_dict()
            return client.reimagine(conds.get("text"))

        supported_params = set(["text"])

        if "text" not in df.columns:
            raise Exception("`text` column has to be given in the query.")

        for col in df.columns:
            if col not in supported_params:
                raise Exception(f"Unknown column {col}. Currently supported parameters for reimagine - {supported_params}")

        client = self._get_clipdrop_client(args)

        return df[df.columns.intersection(supported_params)].apply(generate_reimagine, client=client, axis=1)

    def predict(self, df, args=None):

        args = self.model_storage.json_get('args')

        if args["task"] == "remove_text":
            preds = self._process_remove_text(df, args)
        elif args["task"] == "remove_background":
            preds = self._process_remove_background(df, args)
        elif args["task"] == "sketch_to_image":
            preds = self._process_sketch_to_image(df, args)
        elif args["task"] == "text_to_image":
            preds = self._process_text_to_image(df, args)
        elif args["task"] == "replace_background":
            preds = self._process_replace_background(df, args)
        elif args["task"] == "reimagine":
            preds = self._process_reimagine(df, args)

        result_df = pd.DataFrame()

        result_df['predictions'] = preds

        result_df = result_df.rename(columns={'predictions': args['target']})

        return result_df



import logging
import os
from logging.config import dictConfig

logging_initialized = False


class ColorFormatter(logging.Formatter):

    green = "\x1b[32;20m"
    default = "\x1b[39;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s %(processName)15s %(levelname)-8s %(name)s: %(message)s"

    FORMATS = {
        logging.DEBUG: logging.Formatter(green + format + reset),
        logging.INFO: logging.Formatter(default + format + reset),
        logging.WARNING: logging.Formatter(yellow + format + reset),
        logging.ERROR: logging.Formatter(red + format + reset),
        logging.CRITICAL: logging.Formatter(bold_red + format + reset),
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        return log_fmt.format(record)


def configure_logging():
    mindsdb_level = os.environ.get("MINDSDB_LOG_LEVEL", None)
    if mindsdb_level is not None:
        mindsdb_level = getattr(logging, mindsdb_level)
    else:
        mindsdb_level = logging.INFO

    logging_config = dict(
        version=1,
        formatters={"f": {"()": ColorFormatter}},
        handlers={
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "f",
            }
        },
        loggers={
            "": {  # root logger
                "handlers": ["console"],
                "level": logging.WARNING,
            },
            "__main__": {
                "level": mindsdb_level,
            },
            "mindsdb": {
                "level": mindsdb_level,
            },
            "alembic": {
                "level": mindsdb_level,
            },
        },
    )
    dictConfig(logging_config)


# I would prefer to leave code to use logging.getLogger(), but there are a lot of complicated situations
# in MindsDB with processes being spawned that require logging to be configured again in a lot of cases.
# Using a custom logger-getter like this lets us do that logic here, once.
def getLogger(name=None):
    """
    Get a new logger, configuring logging first if it hasn't been done yet.
    """
    global logging_initialized
    if not logging_initialized:
        configure_logging()
        logging_initialized = True

    return logging.getLogger(name)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        "type": ARG_TYPE.STR,
        "description": "The host name or IP address of the Oracle DB.",
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "The TCP/IP port of the Oracle DB. Must be an integer. Default 1521.",
    },
    sid={
        "type": ARG_TYPE.STR,
        "description": "The site identifier of the Oracle DB. Either sid or service_name should be provided.",
    },
    service_name={
        "type": ARG_TYPE.STR,
        "description": "The name of the Oracle DB service. Either sid or service_name should be provided.",
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate against the Oracle DB.",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user against Oracle DB.",
        "secret": True
    },
    disable_oob={
        "type": ARG_TYPE.BOOL,
        "description": "Disable out-of-band breaks",
    },
    auth_mode={
        "type": ARG_TYPE.STR,
        "description": "Database privilege for connection",
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1",
    port=1521,
    user="admin",
    password="password",
    sid="ORCL",
)



import copy
import importlib
from typing import Dict, Union

import pandas as pd
from langchain.embeddings.base import Embeddings
from pandas import DataFrame

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

logger = log.getLogger(__name__)

# construct the embedding model name to the class mapping
# we try to import all embedding models from langchain_community.embeddings
# for each class, we get a more user friendly name for it
# E.g. OpenAIEmbeddings -> OpenAI
# This is used for the user to select the embedding model
EMBEDDING_MODELS = {}

try:
    module = importlib.import_module("langchain_community.embeddings")
    # iterate __all__ to get all the classes
    for class_name in module.__all__:
        class_ = getattr(module, class_name)
        if not issubclass(class_, Embeddings):
            continue
        # convert the class name to a more user friendly name
        # e.g. OpenAIEmbeddings -> OpenAI
        user_friendly_name = class_name.replace("Embeddings", "")
        EMBEDDING_MODELS[user_friendly_name] = class_name
        EMBEDDING_MODELS[user_friendly_name.lower()] = class_name

except ImportError:
    raise Exception(
        "The langchain is not installed. Please install it with `pip install langchain-community`."
    )


def get_langchain_class(class_name: str) -> Embeddings:
    """Returns the class object of the handler class.

    Args:
        class_name (str): Name of the class

    Returns:
        langchain.embeddings.BaseEmbedding: The class object
    """
    try:
        module = importlib.import_module("langchain_community.embeddings")
        class_ = getattr(module, class_name)
    except ImportError:
        raise Exception(
            "The langchain is not installed. Please install it with `pip install langchain`."
        )
    except AttributeError:
        raise Exception(
            f"Could not find the class {class_name} in langchain_community.embeddings. Please check the class name."
        )
    return class_


def construct_model_from_args(args: Dict) -> Embeddings:
    """
    Deserializes the model from the model storage
    """
    target = args.pop("target", None)
    class_name = args.pop("class", LangchainEmbeddingHandler.DEFAULT_EMBEDDING_CLASS)
    if class_name in EMBEDDING_MODELS:
        logger.info(
            f"Mapping the user friendly name {class_name} to the class name: {EMBEDDING_MODELS[class_name]}"
        )
        class_name = EMBEDDING_MODELS[class_name]
    MODEL_CLASS = get_langchain_class(class_name)
    serialized_dict = copy.deepcopy(args)
    serialized_dict.pop("input_columns", None)
    model = MODEL_CLASS(**serialized_dict)
    if target is not None:
        args["target"] = target
    args["class"] = class_name
    return model


class LangchainEmbeddingHandler(BaseMLEngine):
    """
    Bridge class to connect langchain.embeddings module to mindsDB
    """

    DEFAULT_EMBEDDING_CLASS = "OpenAIEmbeddings"

    def __init__(self, model_storage, engine_storage, **kwargs) -> None:
        super().__init__(model_storage, engine_storage, **kwargs)
        self.generative = True

    def create(
        self,
        target: str,
        df: Union[DataFrame, None] = None,
        args: Union[Dict, None] = None,
    ) -> None:
        # get the class name from the args
        user_args = args.get("using", {})

        # infer the input columns arg if user did not provide it
        # from the columns of the input dataframe if it is provided
        if "input_columns" not in user_args and df is not None:
            # ignore private columns starts with __mindsdb
            # ignore target column in the input dataframe
            user_args["input_columns"] = [
                col
                for col in df.columns.tolist()
                if not col.startswith("__mindsdb") and col != target
            ]
            # unquote the column names -- removing surrounding `
            user_args["input_columns"] = [
                col.strip("`") for col in user_args["input_columns"]
            ]

        elif "input_columns" not in user_args:
            # set as empty list if the input_columns is not provided
            user_args["input_columns"] = []

        # this may raise an exception if
        # the arguments are not sufficient to create such as class
        # due to e.g., lack of API key
        # But the validation logic is handled by langchain and pydantic
        construct_model_from_args(user_args)

        # save the model to the model storage
        target = target or "embeddings"
        user_args[
            "target"
        ] = target  # this is the name of the column to store the embeddings
        self.model_storage.json_set("args", user_args)

    def predict(self, df: DataFrame, args) -> DataFrame:
        # reconstruct the model from the model storage
        user_args = self.model_storage.json_get("args")
        model = construct_model_from_args(user_args)

        # get the target from the model storage
        target = user_args["target"]
        # run the actual embedding vector generation
        # TODO: need a better way to handle this
        # unquote the column names -- removing surrounding `
        cols_dfs = [col.strip("`") for col in df.columns.tolist()]
        df.columns = cols_dfs
        # if input_columns is an empty list, use all the columns
        input_columns = user_args.get("input_columns") or df.columns.tolist()
        # check all the input columns are in the df
        if not all(
            # ignore surrounding ` in the column names when checking
            [col in cols_dfs for col in input_columns]
        ):
            raise Exception(
                f"Input columns {input_columns} not found in the input dataframe. Available columns are {df.columns}"
            )

        # convert each row into a document
        df_texts = df[input_columns].apply(self.row_to_document, axis=1)
        embeddings = model.embed_documents(df_texts.tolist())

        # create a new dataframe with the embeddings
        df_embeddings = df.copy().assign(**{target: embeddings})

        return df_embeddings

    def row_to_document(self, row: pd.Series) -> str:
        """
        Convert a row in the input dataframe into a document

        Default implementation is to concatenate all the columns
        in the form of
        field1: value1\nfield2: value2\n...
        """
        fields = row.index.tolist()
        values = row.values.tolist()
        document = "\n".join(
            [f"{field}: {value}" for field, value in zip(fields, values)]
        )
        return document

    def finetune(
        self, df: Union[DataFrame, None] = None, args: Union[Dict, None] = None
    ) -> None:
        raise NotImplementedError(
            "Finetuning is not supported for langchain embeddings"
        )

    def describe(self, attribute: Union[str, None] = None) -> DataFrame:
        args = self.model_storage.json_get("args")

        if attribute == "args":
            return pd.DataFrame(args.items(), columns=["key", "value"])
        elif attribute == "metadata":
            return pd.DataFrame(
                [
                    ("model_class", self.model_storage.json_get("model_class")),
                ],
                columns=["key", "value"],
            )

        else:
            tables = ("args", "metadata")
            return pd.DataFrame(tables, columns=["tables"])
